# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Augmented Random Search training.

See: https://arxiv.org/pdf/1803.07055.pdf
"""

import functools
import time
from typing import Any, Callable, Dict, Optional, Tuple, Union

from absl import logging
from brax import base
from brax import envs
from brax.training import acting
from brax.training import types
from brax.training.acme import running_statistics
from brax.training.acme import specs
from brax.training.agents.ars import networks as ars_networks
from brax.training.types import Params
from brax.training.types import PRNGKey
from brax.v1 import envs as envs_v1
import flax
import jax
import jax.numpy as jnp
import optax

Metrics = types.Metrics
InferenceParams = Tuple[running_statistics.NestedMeanStd, Params]


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  normalizer_params: running_statistics.RunningStatisticsState
  policy_params: Params
  num_env_steps: int


# TODO: Pass the network as argument.
def train(
    environment: Union[envs_v1.Env, envs.Env],
    num_timesteps: int = 100,
    episode_length: int = 1000,
    action_repeat: int = 1,
    max_devices_per_host: Optional[int] = None,
    number_of_directions: int = 60,
    top_directions: int = 20,
    step_size: float = 0.015,
    num_eval_envs: int = 128,
    exploration_noise_std: float = 0.025,
    seed: int = 0,
    normalize_observations: bool = False,
    num_evals: int = 1,
    reward_shift: float = 0.0,
    network_factory: types.NetworkFactory[
        ars_networks.ARSNetwork
    ] = ars_networks.make_policy_network,
    progress_fn: Callable[[int, Metrics], None] = lambda *args: None,
    eval_env: Optional[envs.Env] = None,
    randomization_fn: Optional[
        Callable[[base.System, jnp.ndarray], Tuple[base.System, base.System]]
    ] = None,
):
  """ARS."""
  top_directions = min(top_directions, number_of_directions)
  num_envs = number_of_directions * 2  # noise + anti noise

  process_count = jax.process_count()
  if process_count > 1:
    raise ValueError('ES is not compatible with multiple hosts, '
                     'please use a single host device.')
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info('Local device count: %d, '
               'devices to be used count: %d', local_device_count,
               local_devices_to_use)

  num_env_steps_between_evals = num_timesteps // num_evals
  next_eval_step = num_timesteps - (num_evals - 1) * num_env_steps_between_evals

  key = jax.random.PRNGKey(seed)
  key, network_key, eval_key, rng_key = jax.random.split(key, 4)

  assert num_envs % local_devices_to_use == 0
  env = environment
  if isinstance(env, envs.Env):
    wrap_for_training = envs.training.wrap
  else:
    wrap_for_training = envs_v1.wrappers.wrap_for_training

  v_randomization_fn = None
  if randomization_fn is not None:
    v_randomization_fn = functools.partial(
        randomization_fn,
        rng=jax.random.split(rng_key, num_envs // local_devices_to_use),
    )
  env = wrap_for_training(
      env,
      episode_length=episode_length,
      action_repeat=action_repeat,
      randomization_fn=v_randomization_fn,
  )

  obs_size = env.observation_size

  normalize_fn = lambda x, y: x
  if normalize_observations:
    normalize_fn = running_statistics.normalize
  ars_network = network_factory(
      observation_size=obs_size,
      action_size=env.action_size,
      preprocess_observations_fn=normalize_fn)
  make_policy = ars_networks.make_inference_fn(ars_network)

  vmapped_policy = jax.vmap(ars_network.apply, in_axes=(None, 0, 0))

  def run_step(carry, unused_target_t):
    (env_state, policy_params, cumulative_reward, active_episode,
     normalizer_params) = carry
    obs = env_state.obs
    actions = vmapped_policy(normalizer_params, policy_params, obs)
    nstate = env.step(env_state, actions)
    cumulative_reward = cumulative_reward + (nstate.reward -
                                             reward_shift) * active_episode
    new_active_episode = active_episode * (1 - nstate.done)
    return (nstate, policy_params, cumulative_reward, new_active_episode,
            normalizer_params), (env_state.obs, active_episode)

  def run_episode(normalizer_params: running_statistics.NestedMeanStd,
                  params: Params, key: PRNGKey):
    reset_keys = jax.random.split(key, num_envs // local_devices_to_use)
    first_env_states = env.reset(reset_keys)
    cumulative_reward = first_env_states.reward
    active_episode = jnp.ones_like(cumulative_reward)
    (_, _, cumulative_reward, _, _), (obs, obs_weights) = jax.lax.scan(
        run_step, (first_env_states, params, cumulative_reward, active_episode,
                   normalizer_params), (),
        length=episode_length // action_repeat)
    return cumulative_reward, obs, obs_weights

  def add_noise(params: Params, key: PRNGKey) -> Tuple[Params, Params, Params]:
    num_vars = len(jax.tree_util.tree_leaves(params))
    treedef = jax.tree_util.tree_structure(params)
    all_keys = jax.random.split(key, num=num_vars)
    noise = jax.tree_util.tree_map(
        lambda g, k: jax.random.normal(k, shape=g.shape, dtype=g.dtype), params,
        jax.tree_util.tree_unflatten(treedef, all_keys))
    params_with_noise = jax.tree_util.tree_map(
        lambda g, n: g + n * exploration_noise_std, params, noise
    )
    params_with_anti_noise = jax.tree_util.tree_map(
        lambda g, n: g - n * exploration_noise_std, params, noise)
    return params_with_noise, params_with_anti_noise, noise

  prun_episode = jax.pmap(run_episode, in_axes=(None, 0, 0))

  @jax.jit
  def training_epoch(training_state: TrainingState,
                     key: PRNGKey) -> Tuple[TrainingState, Metrics]:
    params = jax.tree_util.tree_map(
        lambda x: jnp.repeat(
            jnp.expand_dims(x, axis=0), number_of_directions, axis=0
        ),
        training_state.policy_params,
    )
    key, key_noise, key_es_eval = jax.random.split(key, 3)
    # generate perturbations
    params_with_noise, params_with_anti_noise, noise = add_noise(
        params, key_noise)

    pparams = jax.tree_util.tree_map(
        lambda a, b: jnp.concatenate([a, b], axis=0),
        params_with_noise,
        params_with_anti_noise,
    )

    pparams = jax.tree_util.tree_map(
        lambda x: jnp.reshape(x, (local_devices_to_use, -1) + x.shape[1:]),
        pparams)

    key_es_eval = jax.random.split(key_es_eval, local_devices_to_use)
    eval_scores, obs, obs_weights = prun_episode(
        training_state.normalizer_params, pparams, key_es_eval)

    obs = jnp.reshape(obs, (-1,) + obs.shape[2:])
    obs_weights = jnp.reshape(obs_weights, (-1,) + obs_weights.shape[2:])

    normalizer_params = running_statistics.update(
        training_state.normalizer_params, obs, weights=obs_weights)

    eval_scores = jnp.reshape(eval_scores, [-1])

    reward_plus, reward_minus = jnp.split(eval_scores, 2, axis=0)
    reward_max = jnp.maximum(reward_plus, reward_minus)
    reward_rank = jnp.argsort(jnp.argsort(-reward_max))
    reward_weight = jnp.where(reward_rank < top_directions, 1, 0)
    reward_weight_double = jnp.concatenate([reward_weight, reward_weight],
                                           axis=0)
    reward_std = jnp.std(eval_scores, where=reward_weight_double)
    reward_std += (reward_std == 0.0) * 1e-6

    noise = jax.tree_util.tree_map(
        lambda x: jnp.sum(
            jnp.transpose(
                jnp.transpose(x) * reward_weight * (reward_plus - reward_minus)
            ),
            axis=0,
        ),
        noise,
    )

    policy_params = jax.tree_util.tree_map(
        lambda x, y: x + step_size * y / (top_directions * reward_std),
        training_state.policy_params, noise)

    num_env_steps = training_state.num_env_steps + jnp.sum(
        obs_weights, dtype=jnp.int32) * action_repeat

    metrics = {
        'params_norm': optax.global_norm(policy_params),
        'eval_scores_mean': jnp.mean(eval_scores),
        'eval_scores_std': jnp.std(eval_scores),
        'weights': jnp.mean(reward_weight),
    }
    return (TrainingState(  # type: ignore  # jnp-type
        normalizer_params=normalizer_params,
        policy_params=policy_params,
        num_env_steps=num_env_steps), metrics)

  training_walltime = 0.

  # Note that this is NOT a pure jittable method.
  def training_epoch_with_timing(training_state: TrainingState,
                                 key: PRNGKey) -> Tuple[TrainingState, Metrics]:
    nonlocal training_walltime
    t = time.time()
    (training_state, metrics) = training_epoch(training_state, key)
    metrics = jax.tree_util.tree_map(jnp.mean, metrics)
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), metrics)

    epoch_training_time = time.time() - t
    training_walltime += epoch_training_time
    sps = (num_envs * episode_length) / epoch_training_time
    metrics = {
        'training/sps': sps,
        'training/walltime': training_walltime,
        **{f'training/{name}': value for name, value in metrics.items()}
    }
    return training_state, metrics  # pytype: disable=bad-return-type  # py311-upgrade

  normalizer_params = running_statistics.init_state(
      specs.Array((obs_size,), jnp.dtype('float32')))
  policy_params = ars_network.init(network_key)
  training_state = TrainingState(
      normalizer_params=normalizer_params,
      policy_params=policy_params,
      num_env_steps=0)

  if not eval_env:
    eval_env = environment

  if randomization_fn is not None:
    v_randomization_fn = functools.partial(
        randomization_fn, rng=jax.random.split(eval_key, num_eval_envs)
    )
  eval_env = wrap_for_training(
      eval_env,
      episode_length=episode_length,
      action_repeat=action_repeat,
      randomization_fn=v_randomization_fn,
  )

  # Evaluator function
  evaluator = acting.Evaluator(
      eval_env,
      make_policy,
      num_eval_envs=num_eval_envs,
      episode_length=episode_length,
      action_repeat=action_repeat,
      key=eval_key)

  while training_state.num_env_steps < num_timesteps:
    # optimization
    key, epoch_key = jax.random.split(key)
    training_state, training_metrics = training_epoch_with_timing(
        training_state, epoch_key)

    if training_state.num_env_steps >= next_eval_step:
      # Run evals.
      metrics = evaluator.run_evaluation(
          (training_state.normalizer_params, training_state.policy_params),
          training_metrics)
      logging.info(metrics)
      progress_fn(int(training_state.num_env_steps), metrics)
      next_eval_step += num_env_steps_between_evals

  total_steps = training_state.num_env_steps
  assert total_steps >= num_timesteps

  logging.info('total steps: %s', total_steps)
  params = training_state.normalizer_params, training_state.policy_params
  return (make_policy, params, metrics)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO networks."""

from typing import Sequence, Tuple

from brax.training import distribution
from brax.training import networks
from brax.training import types
from brax.training.types import PRNGKey
import flax
from flax import linen


@flax.struct.dataclass
class PPONetworks:
  policy_network: networks.FeedForwardNetwork
  value_network: networks.FeedForwardNetwork
  parametric_action_distribution: distribution.ParametricDistribution


def make_inference_fn(ppo_networks: PPONetworks):
  """Creates params and inference function for the PPO agent."""

  def make_policy(params: types.PolicyParams,
                  deterministic: bool = False) -> types.Policy:
    policy_network = ppo_networks.policy_network
    parametric_action_distribution = ppo_networks.parametric_action_distribution

    def policy(observations: types.Observation,
               key_sample: PRNGKey) -> Tuple[types.Action, types.Extra]:
      logits = policy_network.apply(*params, observations)
      if deterministic:
        return ppo_networks.parametric_action_distribution.mode(logits), {}
      raw_actions = parametric_action_distribution.sample_no_postprocessing(
          logits, key_sample)
      log_prob = parametric_action_distribution.log_prob(logits, raw_actions)
      postprocessed_actions = parametric_action_distribution.postprocess(
          raw_actions)
      return postprocessed_actions, {
          'log_prob': log_prob,
          'raw_action': raw_actions
      }

    return policy

  return make_policy


def make_ppo_networks(
    observation_size: int,
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types
    .identity_observation_preprocessor,
    policy_hidden_layer_sizes: Sequence[int] = (32,) * 4,
    value_hidden_layer_sizes: Sequence[int] = (256,) * 5,
    activation: networks.ActivationFn = linen.swish) -> PPONetworks:
  """Make PPO networks with preprocessor."""
  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size)
  policy_network = networks.make_policy_network(
      parametric_action_distribution.param_size,
      observation_size,
      preprocess_observations_fn=preprocess_observations_fn,
      hidden_layer_sizes=policy_hidden_layer_sizes,
      activation=activation)
  value_network = networks.make_value_network(
      observation_size,
      preprocess_observations_fn=preprocess_observations_fn,
      hidden_layer_sizes=value_hidden_layer_sizes,
      activation=activation)

  return PPONetworks(
      policy_network=policy_network,
      value_network=value_network,
      parametric_action_distribution=parametric_action_distribution)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Trains an ant to run in the +x direction."""

from brax import base
from brax.envs.base import PipelineEnv, State
from brax.io import mjcf
from etils import epath
import jax
from jax import numpy as jp


class Swimmer(PipelineEnv):



  # pyformat: disable
  """
  ### Description

  This environment corresponds to the Swimmer environment described in Rémi Coulom's PhD thesis
  ["Reinforcement Learning Using Neural Networks, with Applications to Motor Control"](https://tel.archives-ouvertes.fr/tel-00003985/document).
  The environment aims to increase the number of independent state and control
  variables as compared to the classic control environments. The swimmers
  consist of three or more segments ('***links***') and one less articulation
  joints ('***rotors***') - one rotor joint connecting exactly two links to
  form a linear chain. The swimmer is suspended in a two dimensional pool and
  always starts in the same position (subject to some deviation drawn from a
  uniform distribution), and the goal is to move as fast as possible towards
  the right by applying torque on the rotors and using the fluids friction.

  ### Notes

  The problem parameters are:
  * *n*: number of body parts
  * *m<sub>i</sub>*: mass of part *i* (*i* ∈ {1...n})
  * *l<sub>i</sub>*: length of part *i* (*i* ∈ {1...n})
  * *k*: viscous-friction coefficient
  While the default environment has *n* = 3, *l<sub>i</sub>* = 0.1,
  and *k* = 0.1, it is possible to tweak the MuJoCo XML files to increase the
  number of links, or to tweak any of the parameters.

  ### Action Space

  The agent takes a 2-element vector for actions. The action space is a
  continuous `(action, action)` in `[-1, 1]`, where `action` represents the
  numerical torques applied between *links*.
  | Num | Action                             | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Torque applied on the first rotor  | -1          | 1           | rot2                           | hinge | torque (N m) |
  | 1   | Torque applied on the second rotor | -1          | 1           | rot3                           | hinge | torque (N m) |

  ### Observation Space

  The state space consists of:
  * A<sub>0</sub>: position of first point
  * θ<sub>i</sub>: angle of part *i* with respect to the *x* axis
  * A<sub>0</sub>, θ<sub>i</sub>: their derivatives with respect to time (velocity and angular velocity)

  The observation is a `ndarray` with shape `(8,)` where the elements correspond to the following:

  | Num | Observation                          | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|--------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | angle of the front tip               | -Inf | Inf | rot                            | hinge | angle (rad)              |
  | 1   | angle of the second rotor            | -Inf | Inf | rot2                           | hinge | angle (rad)              |
  | 2   | angle of the second rotor            | -Inf | Inf | rot3                           | hinge | angle (rad)              |
  | 3   | velocity of the tip along the x-axis | -Inf | Inf | slider1                        | slide | velocity (m/s)           |
  | 4   | velocity of the tip along the y-axis | -Inf | Inf | slider2                        | slide | velocity (m/s)           |
  | 5   | angular velocity of front tip        | -Inf | Inf | rot                            | hinge | angular velocity (rad/s) |
  | 6   | angular velocity of second rotor     | -Inf | Inf | rot2                           | hinge | angular velocity (rad/s) |
  | 7   | angular velocity of third rotor      | -Inf | Inf | rot3                           | hinge | angular velocity (rad/s) |

  ### Rewards

  The reward consists of two parts:
  - *reward_fwd*: A reward of moving forward which is measured as
    *(x-coordinate before action - x-coordinate after action) / dt*. *dt* is the
    time between actions - the default *dt = 0.04*. This reward would be positive
    if the swimmer swims right as desired.
  - *reward_control*: A negative reward for penalising the swimmer if it takes
    actions that are too large. It is measured as *-coefficient x
    sum(action<sup>2</sup>)* where *coefficient* is a parameter set for the
    control and has a default value of 0.0001

  ### Starting State

  All observations start in state (0,0,0,0,0,0,0,0) with a Uniform noise in the
  range of [-0.1, 0.1], added to the initial state for stochasticity.

  ### Episode Termination

  The episode terminates when the episode length is greater than 1000.
  """
  # pyformat: enable


  def __init__(self,
               forward_reward_weight=1.0,
               ctrl_cost_weight=1e-4,
               reset_noise_scale=0.1,
               exclude_current_positions_from_observation=True,
               backend='generalized',
               **kwargs):
    path = epath.resource_path('brax') / 'envs/assets/swimmer.xml'
    sys = mjcf.load(path)

    n_frames = 4

    if backend not in ['generalized']:
      raise ValueError(f'Unsupported backend: {backend}.')

    kwargs['n_frames'] = kwargs.get('n_frames', n_frames)

    super().__init__(sys=sys, backend=backend, **kwargs)

    self._forward_reward_weight = forward_reward_weight
    self._ctrl_cost_weight = ctrl_cost_weight
    self._reset_noise_scale = reset_noise_scale
    self._exclude_current_positions_from_observation = (
        exclude_current_positions_from_observation)

  def reset(self, rng: jax.Array) -> State:
    rng, rng1, rng2 = jax.random.split(rng, 3)
    qpos = self.sys.init_q + self._noise(rng1)
    qvel = self._noise(rng2)
    pipeline_state = self.pipeline_init(qpos, qvel)
    obs = self._get_obs(pipeline_state)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'reward_fwd': zero,
        'reward_ctrl': zero,
        'x_position': zero,
        'y_position': zero,
        'distance_from_origin': zero,
        'x_velocity': zero,
        'y_velocity': zero,
        'forward_reward': zero,
    }
    return State(pipeline_state, obs, reward, done, metrics)

  def step(self, state: State, action: jax.Array) -> State:
    pipeline_state0 = state.pipeline_state
    pipeline_state = self.pipeline_step(pipeline_state0, action)

    if pipeline_state0 is None:
      raise AssertionError(
          'Cannot compute rewards with pipeline_state0 as Nonetype.')

    xy_position = pipeline_state.q[:2]

    x_velocity = (xy_position[0] - pipeline_state0.q[0]) / self.dt
    y_velocity = (xy_position[1] - pipeline_state0.q[1]) / self.dt

    forward_reward = self._forward_reward_weight * x_velocity
    ctrl_cost = self._ctrl_cost_weight * jp.sum(jp.square(action))

    obs = self._get_obs(pipeline_state)
    reward = forward_reward - ctrl_cost
    state.metrics.update(
        reward_fwd=forward_reward,
        reward_ctrl=-ctrl_cost,
        x_position=xy_position[0],
        y_position=xy_position[1],
        distance_from_origin=jp.linalg.norm(xy_position),
        x_velocity=x_velocity,
        y_velocity=y_velocity,
    )

    return state.replace(pipeline_state=pipeline_state, obs=obs, reward=reward)

  def _get_obs(self, pipeline_state: base.State) -> jax.Array:
    """Observe swimmer body position and velocities."""
    joint_angle = pipeline_state.q
    joint_vel = pipeline_state.qd
    if self._exclude_current_positions_from_observation:
      joint_angle = joint_angle[2:]
    return jp.concatenate((joint_angle, joint_vel))

  def _noise(self, rng, dim=5):
    low, hi = -self._reset_noise_scale, self._reset_noise_scale
    return jax.random.uniform(rng, (dim,), minval=low, maxval=hi)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Base types for spring pipeline."""

from brax import base
from brax.base import Motion, Transform
from flax import struct
import jax
from jax import numpy as jp


@struct.dataclass
class State(base.State):
  """Dynamic state that changes after every step.

  Attributes:
    x_i: (num_links,) link center of mass position in world frame
    xd_i: (num_links,) link center of mass velocity in world frame
    j: link position in joint frame
    jd: link motion in joint frame
    a_p: joint parent anchor in world frame
    a_c: joint child anchor in world frame
    i_inv: link inverse inertia
    mass: link mass
  """
  x_i: Transform
  xd_i: Motion
  j: Transform
  jd: Motion
  a_p: Transform
  a_c: Transform
  i_inv: jax.Array
  mass: jax.Array



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for training wrappers."""
import functools

from absl.testing import absltest
from brax import envs
from brax.envs.wrappers import training
import jax
import jax.numpy as jp
import numpy as np


class TrainingTest(absltest.TestCase):

  def test_domain_randomization_wrapper(self):
    def rand(sys, rng):
      @jax.vmap
      def get_offset(rng):
        offset = jax.random.uniform(rng, shape=(3,), minval=-0.1, maxval=0.1)
        pos = sys.link.transform.pos.at[0].set(offset)
        return pos

      sys_v = sys.tree_replace({'link.inertia.transform.pos': get_offset(rng)})
      in_axes = jax.tree.map(lambda x: None, sys)
      in_axes = in_axes.tree_replace({'link.inertia.transform.pos': 0})
      return sys_v, in_axes

    env = envs.create('ant')
    rng = jax.random.PRNGKey(0)
    rng = jax.random.split(rng, 256)
    env = training.wrap(
        env,
        episode_length=200,
        randomization_fn=functools.partial(rand, rng=rng),
    )

    # set the same key across the batch for env.reset so that only the
    # randomization wrapper creates variability in the env.step
    key = jp.zeros((256, 2), dtype=jp.uint32)
    state = jax.jit(env.reset)(key)
    self.assertEqual(state.pipeline_state.q[:, 0].shape[0], 256)
    self.assertEqual(np.unique(state.pipeline_state.q[:, 0]).shape[0], 1)

    # test that the DomainRandomizationWrapper creates variability in env.step
    state = jax.jit(env.step)(state, jp.zeros((256, env.sys.act_size())))
    self.assertEqual(state.pipeline_state.q[:, 0].shape[0], 256)
    self.assertEqual(np.unique(state.pipeline_state.q[:, 0]).shape[0], 256)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logging utilities."""






import collections
import copy
import csv
import logging
import os
import pprint
import sys
import time
from brax.v1.io import file as io_file
import numpy as np

_tabulators = {}
_timers = {}
_record_time = False


def save_config(output_path, config, verbose=False):
  io_file.MakeDirs(os.path.dirname(output_path))
  config_str = pprint.pformat(config, indent=2)
  with io_file.File(output_path, 'w') as f:
    f.write(config_str)
  if verbose:
    print(f'Saved {output_path}')


def load_config(path, verbose=False):
  with io_file.File(path, 'r') as f:
    config_str = f.read()
  config = eval(config_str)
  if verbose:
    print(f'Loaded {path}')
  return config


class Graph(object):
  """Visualize data in dynamic graphs."""

  def __init__(
      self,
      max_length=100,
  ):
    import pyqtgraph as pg
    from pyqtgraph.Qt import QtGui
    self.max_length = max_length
    self.app = QtGui.QApplication([])
    self.win = pg.GraphicsWindow()
    self.ps = {}
    self.curves = {}
    self.dats = {}

  def add_plot(self, key):
    if key not in self.ps:
      self.ps[key] = self.win.addPlot(colspan=2)
      self.ps[key].setLabel(axis='top', text=key)
      self.win.nextRow()
      self.curves[key] = self.ps[key].plot()
      self.dats[key] = collections.deque()

  def update(self, **values):
    for key, value in values.items():
      self.add_plot(key)
      if len(self.dats[key]) > self.max_length:
        self.dats[key].popleft()
      self.dats[key].append(value)
      self.curves[key].setData(self.dats[key])
    self.app.processEvents()


class Tabulator(object):
  """Tabulate data and incrementally format into a csv."""

  def __init__(self, output_path=None, append=True, cast_type=None):
    self._output_path = output_path
    self._cast_type = cast_type
    self._curr_values = collections.OrderedDict()
    self._history_values = collections.OrderedDict()
    self._curr_counts = {}
    if append and output_path and io_file.Exists(self._output_path):
      self.finalize_from_file()
    else:
      self._finalized = False

  def finalize_from_file(self):
    data = parse_csv(self._output_path)
    self._history_values = data
    for key, value in data.items():
      self._curr_values[key] = value[-1]
    self._finalized = True

  def get_statistics(self, indices=None, stat='mean', tag='', key_filter=None):
    """Get statistics (average, max, min) values in the table."""
    ret = {}
    for key, values in self._history_values.items():
      if key_filter and not key_filter(key):
        continue
      target_values = np.array(values)
      if indices is None:
        pass
      elif isinstance(indices, (int, tuple, list, np.ndarray)):
        if isinstance(indices, int):
          indices = [indices]
        target_values = target_values[indices]
      elif isinstance(indices, str) and ':' in indices:
        first_index, second_index = [
            int(s) if s else None for s in indices.split(':', 1)
        ]
        target_values = target_values[first_index:second_index]
      else:
        raise NotImplementedError(indices, type(indices))
      if tag:
        key += tag
      if stat == 'mean':
        ret[key] = np.mean(target_values, axis=0)
      elif stat == 'max':
        ret[key] = np.max(target_values, axis=0)
      elif stat == 'min':
        ret[key] = np.min(target_values, axis=0)
      else:
        raise NotImplementedError(stat)
    return ret

  def get_last(self):
    return self.get_statistics(indices=-1)

  def get_curr(self):
    return self._curr_values

  def add(self, accumulate=False, **entries):
    """Add an entry of data."""
    for key, value in sorted(entries.items()):
      if key not in self._history_values:
        assert not self._finalized, ('Cannot add a new key {} once tabulator is'
                                     ' finalized.').format(key)
        self._history_values[key] = []
      value = copy.deepcopy(value)
      if accumulate:
        value += self._curr_values.get(key, 0.0)
        self._curr_counts[key] = self._curr_counts.get(key, 0) + 1
      value = self.cast(value)
      self._curr_values[key] = value

  def cast(self, value):
    if self._cast_type:
      try:
        value = self._cast_type(value)
      except TypeError as e:
        raise TypeError('{}: Failed to cast {} as {}'.format(
            e, value, self._cast_type))
    return value

  def finalize(self):
    output_dir = os.path.dirname(self._output_path)
    if output_dir:
      io_file.MakeDirs(output_dir)
    with io_file.File(self._output_path, 'w') as f:
      writer = csv.writer(f)
      writer.writerow(self._history_values.keys())
    self._finalized = True

  def dump(self, output_path=None, average=True):
    """Dump to a csv file."""
    output_path = output_path or self._output_path
    if not self._curr_values:
      return  # empty
    if output_path:
      if not self._finalized:
        self.finalize()  # finalize
      with io_file.File(output_path, 'a') as f:
        writer = csv.writer(f)
        writer.writerow(self._curr_values.values())
    for key, value in self._history_values.items():
      v = copy.deepcopy(self._curr_values[key])
      if average:
        v /= self._curr_counts.get(key, 1.0)
        value = self.cast(value)
      value.append(v)
    self._curr_counts = {}
    self._curr_values = {k: 0.0 for k in self._curr_values}


def parse_csv(filename: str, verbose: bool = False):
  """Parse a csv file."""
  with io_file.File(filename, 'r') as f:
    csv_data = np.genfromtxt(f, delimiter=',', names=True, deletechars='')
  data = collections.OrderedDict()
  try:
    for i, key in enumerate(csv_data.dtype.names):
      data[key] = [d[i] for d in csv_data]
  except TypeError:
    # 0-D array errors out
    for key in csv_data.dtype.names:
      data[key] = np.array([csv_data[key]])
  if verbose:
    print(f'Loaded len={len(list(data.values())[0])}, '
          f'keys={sorted(list(data.keys()))} from {filename}')
  return data


def parse_csv_parallel(filenames, n_threads=1):
  import multiprocessing
  with multiprocessing.pool.ThreadPool(n_threads) as pool:
    jobs = {
        filename: pool.apply_async(parse_csv, [filename], error_callback=print)
        for filename in filenames
    }
    data = {key: value.get() for key, value in jobs.items()}
  return data


def timeit():
  global _record_time
  _record_time = True


def tic(name):
  global _timers
  _timers[name] = time.time()


def toc(name, indent=0):
  global _timers, _record_time
  assert name in _timers
  dt = time.time() - _timers[name]
  del _timers[name]
  if _record_time:
    print('{}[{}] runtime: {}s'.format(''.join(['\t'] * indent), name, dt))
  return dt


def get_level(name):
  """Get level."""
  level = 'info'  # default level
  os_level = os.getenv('LEVEL')
  if os_level is not None:
    if ',' in os_level:
      os_levels = os_level.split(',')
      if name in os_levels[1:]:
        level = os_levels[0]
    else:
      level = os_level
  return level


class LoggerWrapper(object):
  """LoggerWrapper."""

  def __init__(self, logger, name):
    self.logger = logger
    self.name = name

  def format(self, content='', name=None, **kwargs):
    """Format content to str."""
    if name is None:
      name = self.name
    else:
      name = self.name + ':' + name
    s = '[{}]'.format(name)
    if content:
      s += ' ' + pprint.pformat(content)
    if kwargs:
      s += ' ' + pprint.pformat(kwargs)
    return s

  def add_name(self, name):
    self.name = ':'.join((self.name, name))

  def pop_name(self):
    self.name = ':'.join(self.name.split(':')[:-1])

  def debug(self, content='', name=None, **kwargs):
    level = get_level(self.name)
    if level in ('debug',):
      self.logger.debug(self.format(content=content, name=name, **kwargs))

  def info(self, content='', name=None, **kwargs):
    self.logger.info(self.format(content=content, name=name, **kwargs))


def get_logger(level=None, name=__name__):
  """Get logger.

  If `level` is not specified, it consults os.getenv('LEVEL').
      e.g. LEVEL=debug: print all debug messages.
           LEVEL=debug,name1,name2: print all debug messages,
              only for loggers with `name1` or `name2`,
              and use default level (`info`) for others.

  Args:
      level: a string, e.g. 'info', 'debug', 'error'.
      name: a string, identifier for logger.

  Returns:
      A logging.logger object.
  """
  name = name.split('.')[-1]  # base name
  if level is None:
    level = get_level(name)
  logger = logging.getLogger(name)
  out_hdlr = logging.StreamHandler(sys.stdout)
  out_hdlr.setFormatter(
      logging.Formatter('[{}] %(asctime)s %(message)s'.format(name)))
  out_hdlr.setLevel(getattr(logging, level.upper()))
  logger.addHandler(out_hdlr)
  logger.setLevel(getattr(logging, level.upper()))
  return LoggerWrapper(logger, name=name)


def get_tabulator(name=__name__, **kwargs):
  """Get a tabulator."""
  global _tabulators
  if name not in _tabulators:
    _tabulators[name] = Tabulator(**kwargs)
  return _tabulators[name]


if __name__ == '__main__':
  tab = get_tabulator(append=False)
  tab.dump()  # do nothing
  tab.add(a=3, b=2, c=4)
  tab.add(b=4, d=6)
  tab.dump()
  tab.add(a=1, d=4)
  tab.dump()
  tab2 = get_tabulator(append=True)
  tab2.add(a=4, b=1, c=2, d=3)
  tab2.dump()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for com."""
# pylint:disable=g-multiple-import
from absl.testing import absltest
from brax import com
from brax import kinematics
from brax import math
from brax import test_utils
import jax
from jax import numpy as jp
import numpy as np


class ComTest(absltest.TestCase):

  def test_transform(self):
    sys = test_utils.load_fixture('capsule.xml')
    sys = sys.replace(
        link=sys.link.replace(
            inertia=sys.link.inertia.replace(
                transform=sys.link.inertia.transform.replace(
                    pos=jp.array([[0.0, 0.0, -0.1]])
                )
            )
        )
    )

    x, xd = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    x = x.replace(
        pos=jp.array([[1.0, -1.0, 0.3]]),
        rot=jp.array([[0.976442, 0.16639178, -0.13593051, 0.01994515]]),
    )
    xd = xd.replace(
        vel=jp.array([[5.0, 1.0, -1.0]]), ang=jp.array([[1.0, 2.0, -3.0]])
    )

    x_i, xd_i = com.from_world(sys, x, xd)
    self.assertNotAlmostEqual(jp.abs(x_i.pos - x.pos).sum(), 0)
    np.testing.assert_array_almost_equal(x_i.rot, x.rot)
    self.assertNotAlmostEqual(jp.abs(xd_i.vel - xd.vel).sum(), 0)
    np.testing.assert_array_almost_equal(xd_i.ang, xd.ang)
    xp, xdp = com.to_world(sys, x_i, xd_i)
    np.testing.assert_array_almost_equal(x.pos, xp.pos)
    np.testing.assert_array_almost_equal(x.rot, xp.rot)
    np.testing.assert_array_almost_equal(xd.vel, xdp.vel)
    np.testing.assert_array_almost_equal(xd.ang, xdp.ang)

  def test_inv_inertia(self):
    sys = test_utils.load_fixture('capsule.xml')
    sys = sys.replace(
        link=sys.link.replace(
            transform=sys.link.transform.replace(
                rot=math.euler_to_quat(jp.array([0.0, 0.0, 45.0])).reshape(
                    1, -1
                )
            )
        )
    )
    x, _ = kinematics.forward(sys, sys.init_q, jp.zeros(sys.qd_size()))
    x = x.replace(
        rot=math.euler_to_quat(jp.array([45.0, 0.0, 0.0])).reshape(1, -1)
    )

    # get the expected inv inertia
    x_i = x.vmap().do(sys.link.inertia.transform)
    cinr = x_i.replace(pos=jp.zeros_like(x_i.pos)).vmap().do(sys.link.inertia)
    expected = jax.vmap(math.inv_3x3)(cinr.i)

    inv_i = com.inv_inertia(sys, x)
    np.testing.assert_array_almost_equal(expected, inv_i, 1e-6)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""A brax system."""

from typing import Callable, Optional, Sequence, Tuple, Union

from brax.v1 import jumpy as jp
from brax.v1 import math
from brax.v1 import pytree
from brax.v1.experimental.tracing.customize import TracedConfig
from brax.v1.physics import actuators
from brax.v1.physics import bodies
from brax.v1.physics import colliders
from brax.v1.physics import config_pb2
from brax.v1.physics import forces
from brax.v1.physics import geometry
from brax.v1.physics import integrators
from brax.v1.physics import joints
from brax.v1.physics import spring_joints
from brax.v1.physics.base import Info, P, Q, QP, validate_config, vec_to_arr


def _get_contact_info(
    contact: Sequence[geometry.Contact]
) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray]:
  if not contact:
    return jp.array([]), jp.array([]), jp.array([])
  contact = jp.tree_map(jp.concatenate, contact)
  contact = jp.tree_map(lambda *x: jp.concatenate(x), *contact)
  return contact.pos, contact.normal, contact.penetration


@pytree.register
class System:
  """A brax system."""

  __pytree_ignore__ = ('config', 'num_bodies', 'num_joints', 'num_joint_dof',
                       'num_actuators', 'num_forces_dof')

  def __init__(self,
               config: config_pb2.Config,
               resource_paths: Optional[Sequence[str]] = None,
               post_process_fn: Optional[Callable[[config_pb2.Config],
                                                  Union[config_pb2.Config,
                                                        TracedConfig]]] = None):
    config = validate_config(config, resource_paths=resource_paths)
    if post_process_fn:
      config = post_process_fn(config)

    self.config = config
    self.num_actuators = len(config.actuators)
    self.num_joint_dof = sum(len(j.angle_limit) for j in config.joints)
    self.num_bodies = len(config.bodies)
    self.body = bodies.Body(config)
    self.colliders = colliders.get(config, self.body)
    self.num_joints = len(config.joints)
    self.joints = joints.get(config, self.body) + spring_joints.get(
        config, self.body)
    self.actuators = actuators.get(config, self.joints)
    self.forces = forces.get(config, self.body)
    self.num_forces_dof = sum(f.act_index.shape[-1] for f in self.forces)
    self.integrator = integrators.Euler(config)
    zero_p = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
    info = self.info(self.default_qp())
    self.zero_info = Info(
        contact=zero_p,
        joint=zero_p,
        actuator=zero_p,
        contact_pos=jp.zeros_like(info.contact_pos),
        contact_normal=jp.zeros_like(info.contact_normal),
        contact_penetration=-jp.ones_like(info.contact_penetration))

  def default_angle(self, default_index: int = 0) -> jp.ndarray:
    """Returns the default joint angles for the system."""
    if not self.config.joints:
      return jp.array([])

    dofs = {}
    for j in self.config.joints:
      dofs[j.name] = sum([l.min != 0 or l.max != 0 for l in j.angle_limit])
    angles = {}

    # check overrides in config defaults
    if default_index < len(self.config.defaults):
      defaults = self.config.defaults[default_index]
      for ja in defaults.angles:
        angles[ja.name] = vec_to_arr(ja.angle)[:dofs[ja.name]] * jp.pi / 180

    # set remaining joint angles set from angle limits, and add jitter
    for joint in self.config.joints:
      if joint.name not in angles:
        dof = dofs[joint.name]
        angles[joint.name] = jp.array([
            (l.min + l.max) * jp.pi / 360 for l in joint.angle_limit
        ][:dof])

    return jp.concatenate([angles[j.name] for j in self.config.joints])

  def default_qp(self,
                 default_index: int = 0,
                 joint_angle: Optional[jp.ndarray] = None,
                 joint_velocity: Optional[jp.ndarray] = None) -> QP:
    """Returns a default state for the system."""
    qp = QP.zero(shape=(self.num_bodies,))

    # set any default qps from the config
    default = None
    if default_index < len(self.config.defaults):
      default = self.config.defaults[default_index]
      for dqp in default.qps:
        body_i = self.body.index[dqp.name]
        pos = jp.index_update(qp.pos, body_i, vec_to_arr(dqp.pos))
        rot = jp.index_update(qp.rot, body_i,
                              math.euler_to_quat(vec_to_arr(dqp.rot)))
        vel = jp.index_update(qp.vel, body_i, vec_to_arr(dqp.vel))
        ang = jp.index_update(qp.ang, body_i, vec_to_arr(dqp.ang))
        qp = qp.replace(pos=pos, rot=rot, vel=vel, ang=ang)

    # build joints and joint -> array lookup, and order by depth
    if joint_angle is None:
      joint_angle = self.default_angle(default_index)
    if joint_velocity is None:
      joint_velocity = jp.zeros_like(joint_angle)
    joint_idxs = []
    for j in self.config.joints:
      beg = joint_idxs[-1][1][1] if joint_idxs else 0
      dof = sum([l.min != 0 or l.max != 0 for l in j.angle_limit])
      joint_idxs.append((j, (beg, beg + dof)))
    lineage = {j.child: j.parent for j in self.config.joints}
    depth = {}
    for child, parent in lineage.items():
      depth[child] = 1
      while parent in lineage:
        parent = lineage[parent]
        depth[child] += 1
    joint_idxs = sorted(joint_idxs, key=lambda x: depth.get(x[0].parent, 0))
    joint = [j for j, _ in joint_idxs]

    if joint:
      # convert joint_angle and joint_vel to 3dof
      takes = []
      beg = 0
      for j, (beg, end) in joint_idxs:
        arr = list(range(beg, end))
        arr.extend([self.num_joint_dof] * (3 - len(arr)))
        takes.extend(arr)
      takes = jp.array(takes, dtype=int)

      def to_3dof(a):
        a = jp.concatenate([a, jp.array([0.0])])
        a = jp.take(a, takes)
        a = jp.reshape(a, (self.num_joints, 3))
        return a

      joint_angle = to_3dof(joint_angle)
      joint_velocity = to_3dof(joint_velocity)

      # build local rot and ang per joint
      joint_rot = jp.array(
          [math.euler_to_quat(vec_to_arr(j.rotation)) for j in joint])
      joint_ref = jp.array(
          [math.euler_to_quat(vec_to_arr(j.reference_rotation)) for j in joint])

      def local_rot_ang(_, x):
        angles, vels, rot, ref = x
        axes = jp.vmap(math.rotate, [True, False])(jp.eye(3), rot)
        ang = jp.dot(axes.T, vels).T
        rot = ref
        for axis, angle in zip(axes, angles):
          # these are euler intrinsic rotations, so the axes are rotated too:
          axis = math.rotate(axis, rot)
          next_rot = math.quat_rot_axis(axis, angle)
          rot = math.quat_mul(next_rot, rot)
        return (), (rot, ang)

      xs = (joint_angle, joint_velocity, joint_rot, joint_ref)
      _, (local_rot, local_ang) = jp.scan(local_rot_ang, (), xs, len(joint))

      # update qp in depth order
      joint_body = jp.array([
          (self.body.index[j.parent], self.body.index[j.child]) for j in joint
      ])
      joint_off = jp.array([(vec_to_arr(j.parent_offset),
                             vec_to_arr(j.child_offset)) for j in joint])

      def set_qp(carry, x):
        qp, = carry
        (body_p, body_c), (off_p, off_c), local_rot, local_ang = x
        world_rot = math.quat_mul(qp.rot[body_p], local_rot)
        local_pos = off_p - math.rotate(off_c, local_rot)
        world_pos = qp.pos[body_p] + math.rotate(local_pos, qp.rot[body_p])
        world_ang = math.rotate(local_ang, qp.rot[body_p])
        pos = jp.index_update(qp.pos, body_c, world_pos)
        rot = jp.index_update(qp.rot, body_c, world_rot)
        ang = jp.index_update(qp.ang, body_c, world_ang)
        qp = qp.replace(pos=pos, rot=rot, ang=ang)
        return (qp,), ()

      xs = (joint_body, joint_off, local_rot, local_ang)
      (qp,), () = jp.scan(set_qp, (qp,), xs, len(joint))

    # any trees that have no body qp overrides in the config are moved above
    # the xy plane.  this convenience operation may be removed in the future.
    fixed = {j.child for j in joint}
    if default:
      fixed |= {qp.name for qp in default.qps}
    root_idx = {
        b.name: [i]
        for i, b in enumerate(self.config.bodies)
        if b.name not in fixed
    }
    for j in joint:
      parent = j.parent
      while parent in lineage:
        parent = lineage[parent]
      if parent in root_idx:
        root_idx[parent].append(self.body.index[j.child])

    for children in root_idx.values():
      zs = jp.array([
          bodies.min_z(jp.take(qp, c), self.config.bodies[c]) for c in children
      ])
      min_z = jp.amin(zs)
      children = jp.array(children)
      pos = jp.take(qp.pos, children) - min_z * jp.array([0., 0., 1.])
      pos = jp.index_update(qp.pos, children, pos)
      qp = qp.replace(pos=pos)

    return qp

  def step(self, qp: QP, act: jp.ndarray) -> Tuple[QP, Info]:
    """Generic step function.  Overridden with appropriate step at init."""
    step_funs = {'pbd': self._pbd_step, 'legacy_spring': self._spring_step}
    return step_funs[self.config.dynamics_mode](qp, act)

  def info(self, qp: QP) -> Info:
    """Return info about a system state."""
    info_funs = {'pbd': self._pbd_info, 'legacy_spring': self._spring_info}
    return info_funs[self.config.dynamics_mode](qp)

  def _pbd_step(self, qp: QP, act: jp.ndarray) -> Tuple[QP, Info]:
    """Position based dynamics stepping scheme."""

    # Just like XPBD except performs two physics substeps per collision update.

    def substep(carry, _):
      qp, info = carry

      # first substep without collisions
      qprev = qp

      # apply acceleration updates for actuators, and forces
      zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
      zero_q = Q(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 4)))
      dp_a = sum([a.apply(qp, act) for a in self.actuators], zero)
      dp_f = sum([f.apply(qp, act) for f in self.forces], zero)
      dp_j = sum([j.damp(qp) for j in self.joints], zero)
      qp = self.integrator.update(qp, acc_p=dp_a + dp_f + dp_j)

      # apply kinetic step
      qp = self.integrator.kinetic(qp)

      # apply joint position update
      dq_j = sum([j.apply(qp) for j in self.joints], zero_q)
      qp = self.integrator.update(qp, pos_q=dq_j)

      # apply pbd velocity projection
      qp = self.integrator.velocity_projection(qp, qprev)

      qprev = qp
      # second substep with collisions

      # apply acceleration updates for actuators, and forces
      dp_a = sum([a.apply(qp, act) for a in self.actuators], zero)
      dp_f = sum([f.apply(qp, act) for f in self.forces], zero)
      dp_j = sum([j.damp(qp) for j in self.joints], zero)
      qp = self.integrator.update(qp, acc_p=dp_a + dp_f + dp_j)

      # apply kinetic step
      qp = self.integrator.kinetic(qp)

      # apply joint position update
      dq_j = sum([j.apply(qp) for j in self.joints], zero_q)
      qp = self.integrator.update(qp, pos_q=dq_j)

      collide_data = [c.position_apply(qp, qprev) for c in self.colliders]
      dq_c = sum([c[0] for c in collide_data], zero_q)
      dlambda = [c[1] for c in collide_data]
      contact = [c[2] for c in collide_data]
      qp = self.integrator.update(qp, pos_q=dq_c)

      # apply pbd velocity updates
      qp_right_before = qp
      qp = self.integrator.velocity_projection(qp, qprev)
      # apply collision velocity updates
      dp_c = sum([
          c.velocity_apply(qp, dlambda[i], qp_right_before, contact[i])
          for i, c in enumerate(self.colliders)
      ], zero)
      qp = self.integrator.update(qp, vel_p=dp_c)

      info = Info(info.contact + dp_c, info.joint, info.actuator + dp_a,
                  *_get_contact_info(contact))
      return (qp, info), ()

    # update collider statistics for culling
    for c in self.colliders:
      c.cull.update(qp)

    info = self.zero_info
    (qp, info), _ = jp.scan(substep, (qp, info), (), self.config.substeps // 2)
    return qp, info

  def _pbd_info(self, qp: QP) -> Info:
    """Return info about a system state."""
    zero_q = Q(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 4)))
    zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))

    # TODO: sort out a better way to get first-step collider data
    for c in self.colliders:
      c.cull.update(qp)
    collider_data = [c.apply(qp) for c in self.colliders]
    contact = [c[1] for c in collider_data]
    dq_c = sum([c[0] for c in collider_data], zero)
    dq_j = sum([j.apply(qp) for j in self.joints], zero_q)
    info = Info(dq_c, dq_j, zero, *_get_contact_info(contact))
    return info

  def _spring_step(self, qp: QP, act: jp.ndarray) -> Tuple[QP, Info]:
    """Spring-based dynamics stepping scheme."""

    # Resolves actuator forces, joints, and forces at acceleration level, and
    # resolves collisions at velocity level with baumgarte stabilization.

    def substep(carry, _):
      qp, info = carry

      # apply kinetic step
      qp = self.integrator.kinetic(qp)

      # apply acceleration level updates for joints, actuators, and forces
      zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))
      dp_j = sum([j.apply(qp) for j in self.joints], zero)
      dp_a = sum([a.apply(qp, act) for a in self.actuators], zero)
      dp_f = sum([f.apply(qp, act) for f in self.forces], zero)
      qp = self.integrator.update(qp, acc_p=dp_j + dp_a + dp_f)

      # apply velocity level updates for collisions
      collider_data = [c.apply(qp) for c in self.colliders]
      dp_c = sum([c[0] for c in collider_data], zero)
      contact = [c[1] for c in collider_data]
      qp = self.integrator.update(qp, vel_p=dp_c)

      info = Info(info.contact + dp_c, info.joint + dp_j, info.actuator + dp_a,
                  *_get_contact_info(contact))
      return (qp, info), ()

    # update collider statistics for culling
    for c in self.colliders:
      c.cull.update(qp)

    info = self.zero_info
    (qp, info), _ = jp.scan(substep, (qp, info), (), self.config.substeps)
    return qp, info

  def _spring_info(self, qp: QP) -> Info:
    """Return info about a system state."""
    zero = P(jp.zeros((self.num_bodies, 3)), jp.zeros((self.num_bodies, 3)))

    for c in self.colliders:
      c.cull.update(qp)
    collider_data = [c.apply(qp) for c in self.colliders]
    contact = [c[1] for c in collider_data]
    dp_c = sum([c[0] for c in collider_data], zero)
    dp_j = sum([j.apply(qp) for j in self.joints], zero)
    info = Info(dp_c, dp_j, zero, *_get_contact_info(contact))
    return info



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper around a Brax GymWrapper, that converts outputs to PyTorch tensors.

This conversion happens directly on-device, without moving values to the CPU.
"""
from typing import Optional

# NOTE: The following line will emit a warning and raise ImportError if `torch`
# isn't available.
from brax.io import torch
import gym


class TorchWrapper(gym.Wrapper):
  """Wrapper that converts Jax tensors to PyTorch tensors."""

  def __init__(self, env: gym.Env, device: Optional[torch.Device] = None):
    """Creates a gym Env to one that outputs PyTorch tensors."""
    super().__init__(env)
    self.device = device

  def reset(self):
    obs = super().reset()
    return torch.jax_to_torch(obs, device=self.device)

  def step(self, action):
    action = torch.torch_to_jax(action)
    obs, reward, done, info = super().step(action)
    obs = torch.jax_to_torch(obs, device=self.device)
    reward = torch.jax_to_torch(reward, device=self.device)
    done = torch.jax_to_torch(done, device=self.device)
    info = torch.jax_to_torch(info, device=self.device)
    return obs, reward, done, info



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Physics pipeline for fully articulated dynamics and collisiion."""

from typing import Optional
from brax import actuator
from brax import com
from brax import contact
from brax import fluid
from brax import kinematics
from brax.base import Motion, System
from brax.io import mjcf
from brax.spring import collisions
from brax.spring import integrator
from brax.spring import joints
from brax.spring.base import State
import jax


def init(
    sys: System,
    q: jax.Array,
    qd: jax.Array,
    unused_act: Optional[jax.Array] = None,
    debug: bool = False,
) -> State:
  """Initializes physics state.

  Args:
    sys: a brax system
    q: (q_size,) joint angle vector
    qd: (qd_size,) joint velocity vector
    debug: if True, adds contact to the state for debugging

  Returns:
    state: initial physics state
  """
  if sys.mj_model is not None:
    mjcf.validate_model(sys.mj_model)
  # position/velocity level terms
  x, xd = kinematics.forward(sys, q, qd)
  j, jd, a_p, a_c = kinematics.world_to_joint(sys, x, xd)
  x_i, xd_i = com.from_world(sys, x, xd)
  i_inv = com.inv_inertia(sys, x)
  mass = sys.link.inertia.mass ** (1 - sys.spring_mass_scale)

  return State(
      q=q,
      qd=qd,
      x=x,
      xd=xd,
      contact=contact.get(sys, x) if debug else None,
      x_i=x_i,
      xd_i=xd_i,
      j=j,
      jd=jd,
      a_p=a_p,
      a_c=a_c,
      i_inv=i_inv,
      mass=mass,
  )


def step(
    sys: System, state: State, act: jax.Array, debug: bool = False
) -> State:
  """Performs a single physics step using spring-based dynamics.

  Resolves actuator forces, joints, and forces at acceleration level, and
  resolves collisions at velocity level with baumgarte stabilization.

  Args:
    sys: system defining the kinematic tree and other properties
    state: physics state prior to step
    act: (act_size,) actuator input vector
    debug: if True, adds contact to the state for debugging

  Returns:
    x: updated link transform in world frame
    xd: updated link motion in world frame
  """
  # pre-calculate some auxiliary terms used further down
  state = state.replace(i_inv=com.inv_inertia(sys, state.x))

  # calculate acceleration and delta-velocity terms
  tau = actuator.to_tau(sys, act, state.q, state.qd)
  xdd_i = Motion.create(vel=sys.gravity)
  xf_i = joints.resolve(sys, state, tau)
  if sys.enable_fluid:
    inertia = sys.link.inertia.i ** (1 - sys.spring_inertia_scale)
    xf_i += fluid.force(sys, state.x, state.xd, state.mass, inertia)
  xdd_i += Motion(
      ang=jax.vmap(lambda x, y: x @ y)(state.i_inv, xf_i.ang),
      vel=jax.vmap(lambda x, y: x / y)(xf_i.vel, state.mass),
  )

  # semi-implicit euler: apply acceleration update before resolving collisions
  state = state.replace(xd_i=state.xd_i + xdd_i * sys.opt.timestep)
  xdv_i = collisions.resolve(sys, state)

  # now integrate and update position/velocity-level terms
  x_i, xd_i = integrator.integrate(sys, state.x_i, state.xd_i, xdv_i)
  x, xd = com.to_world(sys, x_i, xd_i)
  state = state.replace(x=x, xd=xd, x_i=x_i, xd_i=xd_i)
  j, jd, a_p, a_c = kinematics.world_to_joint(sys, x, xd)
  q, qd = kinematics.inverse(sys, j, jd)
  state = state.replace(
      q=q,
      qd=qd,
      a_p=a_p,
      a_c=a_c,
      j=j,
      jd=jd,
      contact=contact.get(sys, x) if debug else None,
  )

  return state



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.physics.colliders."""

from absl.testing import absltest
import brax.v1 as brax
from brax.v1 import jumpy as jp
from google.protobuf import text_format


class CapsuleCapsuleColliderFnTest(absltest.TestCase):
  """Tests the capsule-capsule collision function."""

  _CONFIG = """
    dt: .01 substeps: 4 friction: 0.6
    gravity { z: -9.8 }
    bodies {
      name: "Capsule1" mass: 1
      colliders { capsule { radius: 0.05 length: 0.88 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule2" mass: 1
      colliders { capsule { radius: 0.05 length: 0.75 } }
      inertia { x: 1 y: 1 z: 1 }
      frozen: { all: true }
    }
    defaults {
      qps {
        name: "Capsule1"
        pos {x: 1.0776836 y: 0.46205616 z: 10.814447 }
        rot {x: 37.72620165 y: 65.59518961 z: 63.17161424}
      }
      qps {
        name: "Capsule2"
        pos {x: 0.9702782 y: 0.45527467 z: 10.704371}
        rot {x: 33.0532861 y: 57.08759044 z: 87.28387177}
      }
    }
  """

  def test_collision(self):
    """Tests that two capsules are colliding."""
    config = text_format.Parse(CapsuleCapsuleColliderFnTest._CONFIG,
                               brax.Config())
    sys = brax.System(config)
    capsules = brax.physics.geometry.Capsule(config.bodies, sys.body)
    capsule1 = jp.take(capsules, 0)
    capsule2 = jp.take(capsules, 1)
    qp1 = jp.take(sys.default_qp(), 0)
    qp2 = jp.take(sys.default_qp(), 1)

    contact = brax.physics.colliders.capsule_capsule(capsule1, capsule2, qp1,
                                                     qp2)

    self.assertGreaterEqual(contact.penetration[0], 0)

  # Parallel capsule collision.
  _CONFIG2 = """
    dt: .01 substeps: 4 friction: 0.6
    gravity { z: -9.8 }
    bodies {
      name: "Capsule1" mass: 1
      colliders { capsule { radius: 0.1 length: 1.0 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule2" mass: 1
      colliders { capsule { radius: 0.1 length: 1.0 } }
      inertia { x: 1 y: 1 z: 1 }
      frozen: { all: true }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {
      qps {
        name: "Capsule1"
        pos {x: 0 y: 0 z: 0.25 }
        rot {x: 0 y: 90 z: 0}
      }
      qps {
        name: "Capsule2"
        pos {x: 0 y: 0 z: 0.1}
        rot {x: 0 y: 90 z: 0}
      }
    }
  """

  def test_parallel_collision(self):
    """Tests that two parallel capsules are colliding at the midpoint."""
    config = text_format.Parse(CapsuleCapsuleColliderFnTest._CONFIG2,
                               brax.Config())
    sys = brax.System(config)
    capsules = brax.physics.geometry.Capsule(config.bodies, sys.body)
    capsule1 = jp.take(capsules, 0)
    capsule2 = jp.take(capsules, 1)
    qp1 = jp.take(sys.default_qp(), 0)
    qp2 = jp.take(sys.default_qp(), 1)

    contact = brax.physics.colliders.capsule_capsule(capsule1, capsule2, qp1,
                                                     qp2)
    self.assertAlmostEqual(contact.penetration[0], 0.05)
    self.assertSequenceAlmostEqual(contact.pos[0], [0., 0., (0.15 + 0.2) / 2.0],
                                   4)

  def test_collision_with_zero_capsule_length(self):
    """Tests collisions with capsules that have zero length (i.e. spheres)."""
    config = text_format.Parse(CapsuleCapsuleColliderFnTest._CONFIG2,
                               brax.Config())
    sys = brax.System(config)
    capsules = brax.physics.geometry.Capsule(config.bodies, sys.body)
    capsule1 = jp.take(capsules, 0)
    capsule2 = jp.take(capsules, 1)
    # Set capsule ends such that the capsule lengths are zero.
    capsule1.end = jp.array([0., 0., 0.])
    capsule2.end = jp.array([0., 0., 0.])
    qp1 = jp.take(sys.default_qp(), 0)
    qp2 = jp.take(sys.default_qp(), 1)

    contact = brax.physics.colliders.capsule_capsule(capsule1, capsule2, qp1,
                                                     qp2)
    self.assertAlmostEqual(contact.penetration[0], 0.05)
    self.assertSequenceAlmostEqual(contact.pos[0], [0., 0., (0.15 + 0.2) / 2.0],
                                   4)


class BoxHeightMapColliderFnTest(absltest.TestCase):
  """Tests the box-heightmap collision function."""
  _CONFIG = """
    dt: .01 substeps: 4 friction: 0.6
    gravity { z: -9.8 }
    bodies {
      name: "box" mass: 1
      colliders { box { halfsize { x: 0.3 y: 0.3 z: 0.3 }}}
      inertia { x: 0.1 y: 0.1 z: 0.1 }
    }
    bodies {
      name: "ground"
      frozen: { all: true }
      colliders {
        heightMap {
          size: 10
          data: [2, 2, 2, 2, 2, 2, 2, 2, 2]
        }
      }
    }
    # Place the box in the first quadrant of the height map.
    defaults { qps { name: "box" pos: {x: 2 y: -2 z: 1.7}}}
  """

  def test_collision_position(self):
    """Tests collision occurs at the correct position on the heightmap."""
    config = text_format.Parse(BoxHeightMapColliderFnTest._CONFIG,
                               brax.Config())
    sys = brax.System(config)
    box_collider = brax.physics.geometry.Box([jp.take(config.bodies, 0)],
                                             sys.body)
    box_collider = jp.take(box_collider, 0)
    heightmap_collider = brax.physics.geometry.HeightMap(
        [jp.take(config.bodies, 1)], sys.body)
    heightmap_collider = jp.take(heightmap_collider, 0)
    box_qp = jp.take(sys.default_qp(), 0)
    hm_qp = jp.take(sys.default_qp(), 1)

    contact = brax.physics.colliders.box_heightmap(box_collider,
                                                   heightmap_collider, box_qp,
                                                   hm_qp)
    self.assertSequenceAlmostEqual(contact.pos[0],
                                   [2 - 0.3, -2 - 0.3, 1.7 - 0.3], 3)
    self.assertGreater(contact.penetration[0], 0.0)


class CapsuleMeshTest(absltest.TestCase):
  """Tests the capsule-mesh collision function."""
  _CONFIG = """
    dt: .01 substeps: 4 friction: 0.6
    gravity { z: -9.8 }
    bodies {
      name: "Capsule1" mass: 1
      colliders { capsule { radius: 0.5 length: 3.14 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Mesh" mass: 1
      colliders { mesh { name: "Mesh1" scale: 1 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    mesh_geometries {
      name: "Mesh1"
      vertices { x: 0.47 y: -0.68 z: 0.24 }
      vertices { x: -1.70 y: 0.75 z: -1.53 }
      vertices { x: 0.01 y: -0.12 z: -0.81 }
      face_normals { x: 0.51 y: 1.45 z: 0.55 }
      faces: [0, 1, 2]
    }
    defaults {
      qps { name: "Capsule1" pos {x: 0} rot { x: 113 y: 74 z: 152 } }
      qps { name: "Mesh" pos {x: 0} }
    }
  """

  def test_collision_position(self):
    """Tests collision occurs between capsule and triangle."""
    config = text_format.Parse(CapsuleMeshTest._CONFIG, brax.Config())
    sys = brax.System(config)

    capsules = brax.physics.geometry.Capsule(config.bodies, sys.body)
    mesh_geoms = {
        mesh_geom.name: mesh_geom for mesh_geom in config.mesh_geometries
    }
    mesh = brax.physics.geometry.Mesh(
        jp.array([jp.take(config.bodies, 1)]), sys.body, mesh_geoms)
    capsule1 = jp.take(capsules, 0)
    mesh = jp.take(mesh, 0)
    qp = sys.default_qp(0)
    contact = brax.physics.colliders.capsule_mesh(capsule1,
                                                  mesh, jp.take(qp, 0),
                                                  jp.take(qp, 1))

    self.assertGreater(contact.penetration[0], 0.0)


class MultiMeshTest(absltest.TestCase):
  """Tests scene with multiple meshes."""
  _CONFIG = """
    dt: .01 substeps: 4 friction: 0.6
    gravity { z: -9.8 }
    bodies {
      name: "Capsule1" mass: 1
      colliders { capsule { radius: 0.5 length: 3.14 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Mesh_1" mass: 1
      colliders { mesh { name: "Mesh1" scale: 1 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    mesh_geometries {
      name: "Mesh1"
      vertices { x: 0.47 y: -0.68 z: 0.24 }
      vertices { x: -1.70 y: 0.75 z: -1.53 }
      vertices { x: 0.01 y: -0.12 z: -0.81 }
      face_normals { x: 0.51 y: 1.45 z: 0.55 }
      faces: [0, 1, 2]
    }
    bodies {
      name: "Mesh_2" mass: 1
      colliders { mesh { name: "Mesh2" scale: 1 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    mesh_geometries {
      name: "Mesh2"
      vertices { x: 0.47 y: -0.68 z: 0.24 }
      vertices { x: -1.70 y: 0.75 z: -1.53 }
      vertices { x: 0.01 y: -0.12 z: -0.81 }
      face_normals { x: 0.51 y: 1.45 z: 0.55 }
      faces: [0, 1, 2]
    }
    defaults {
      qps { name: "Capsule1" pos {x: 0} rot { x: 113 y: 74 z: 152 } }
      qps { name: "Mesh_1" pos {x: 0} }
      qps { name: "Mesh_2" pos {x: 3} }
    }
  """

  def test_mesh_collider_construction(self):
    """Tests collision occurs between capsule and triangle."""
    config = text_format.Parse(MultiMeshTest._CONFIG, brax.Config())
    sys = brax.System(config)
    # builds two collision primitives for each different mesh
    self.assertLen(sys.colliders, 2)
    # checks that collision bodies have correct body indices
    self.assertEqual(sys.colliders[0].cull.col_a.body.idx[0], 0)
    self.assertEqual(sys.colliders[0].cull.col_b.body.idx[0], 1)
    self.assertEqual(sys.colliders[1].cull.col_a.body.idx[0], 0)
    self.assertEqual(sys.colliders[1].cull.col_b.body.idx[0], 2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Objects which specify the input/output spaces of an environment.


This file was taken from acme and modified to simplify dependencies:

https://github.com/deepmind/acme/blob/master/acme/specs.py
"""
import dataclasses
from typing import Tuple

import jax.numpy as jnp


@dataclasses.dataclass(frozen=True)
class Array:
  """Describes a numpy array or scalar shape and dtype.

  Similar to dm_env.specs.Array.
  """
  shape: Tuple[int, ...]
  dtype: jnp.dtype



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for spring physics pipeline."""

from absl.testing import absltest
from brax import com
from brax import kinematics
from brax import test_utils
from brax.base import Transform
from brax.generalized import pipeline as g_pipeline
from brax.spring import pipeline
import jax
from jax import numpy as jp
import numpy as np


class PipelineTest(absltest.TestCase):

  def test_pendulum(self):
    sys = test_utils.load_fixture('triple_pendulum.xml')
    sys = sys.tree_replace({'opt.timestep': 0.0001})
    sys = sys.replace(
        link=sys.link.replace(constraint_stiffness=jp.array([800_000.0] * 3))
    )

    state = pipeline.init(sys, sys.init_q, jp.zeros(sys.qd_size()))
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(10_000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
    x = state.x

    # compare against generalized step
    q, qd = jp.zeros(sys.q_size()), jp.zeros(sys.qd_size())
    state = g_pipeline.init(sys, q, qd)
    j_g_step = jax.jit(g_pipeline.step)
    j_forward = jax.jit(kinematics.forward)
    for _ in range(10_000):
      state = j_g_step(sys, state, jp.zeros(sys.act_size()))
    x_g, _ = j_forward(sys, state.q, state.qd)

    # trajectories should be close after 1 second of simulation
    self.assertLess(jp.linalg.norm(x_g.pos - x.pos), 2e-2)

  def test_universal_pendulum(self):
    sys = test_utils.load_fixture('single_universal_pendulum.xml')

    init_q = jp.array([0.1, 0.2])
    init_qd = jax.random.uniform(jax.random.PRNGKey(0), (2,)) / 10.0

    # set sensible springy defaults
    link = sys.link.replace(constraint_limit_stiffness=jp.array([0.0] * 2))
    link = link.replace(constraint_stiffness=jp.array([100_000.0] * 2))
    link = link.replace(constraint_ang_damping=jp.array([0.0] * 2))
    link = link.replace(constraint_vel_damping=jp.array([200.0] * 2))
    sys = sys.replace(link=link)
    sys = sys.replace(ang_damping=0.0)
    sys = sys.tree_replace({'opt.timestep': 0.001})
    sys = sys.replace(solver_iterations=500)

    state = pipeline.init(sys, init_q, init_qd)
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
    x = state.x

    # compare against generalized step
    state = g_pipeline.init(sys, init_q, init_qd)
    j_g_step = jax.jit(g_pipeline.step)
    j_forward = jax.jit(kinematics.forward)
    for _ in range(1000):
      state = j_g_step(sys, state, jp.zeros(sys.qd_size()))
    x_g, _ = j_forward(sys, state.q, state.qd)

    # trajectories should be close after 1 second of simulation
    self.assertLess(jp.linalg.norm(x_g.rot - x.rot), 1.51e-2)

  def test_spherical_pendulum(self):
    sys = test_utils.load_fixture('single_spherical_pendulum.xml')

    init_q = jp.array([0.1, 0.2, 0.3])
    init_qd = jax.random.uniform(jax.random.PRNGKey(0), (3,)) / 10.0

    # set sensible springy defaults
    link = sys.link.replace(constraint_limit_stiffness=jp.array([0.0] * 3))
    link = link.replace(constraint_stiffness=jp.array([100_000.0] * 3))
    link = link.replace(constraint_ang_damping=jp.array([0.0] * 3))
    link = link.replace(constraint_vel_damping=jp.array([200.0] * 3))
    sys = sys.replace(link=link)
    sys = sys.replace(ang_damping=0.0)
    sys = sys.tree_replace({'opt.timestep': 0.001})
    sys = sys.replace(solver_iterations=500)

    state = pipeline.init(sys, init_q, init_qd)
    # the qd calculation for pbd/spring doesn't match generalized, so we get xd
    # from generalized and plug it back into pbd
    # TODO: remove this xd override once kinematics.forward is fixed
    state_g = g_pipeline.init(sys, init_q, init_qd)
    off = state_g.x.pos - state_g.root_com
    xd = Transform.create(pos=off).vmap().do(state_g.cd)
    state = state.replace(xd=xd, xd_i=com.from_world(sys, state.x, xd)[1])

    j_spring_step = jax.jit(pipeline.step)
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
    x = state.x

    # compare against generalized step
    state = state_g
    j_g_step = jax.jit(g_pipeline.step)
    j_forward = jax.jit(kinematics.forward)
    for _ in range(1000):
      state = j_g_step(sys, state, jp.zeros(sys.act_size()))
    x_g, _ = j_forward(sys, state.q, state.qd)

    # trajectories should be close after 1 second of simulation
    self.assertLess(jp.linalg.norm(x_g.rot - x.rot), 1e-2)

  def test_prismatic_joint(self):
    # tests launching a mass along a slide joint in the [1,1,1] direction
    sys = test_utils.load_fixture('single_prismatic.xml')

    init_q = jp.array([0.9])
    init_qd = jp.array([1.0])

    # set sensible springy defaults
    link = sys.link.replace(constraint_limit_stiffness=jp.array([10_000.0]))
    link = link.replace(constraint_stiffness=jp.array([10_000.0]))
    link = link.replace(constraint_ang_damping=jp.array([0.0]))
    link = link.replace(constraint_vel_damping=jp.array([200.0]))
    sys = sys.replace(link=link)
    sys = sys.replace(ang_damping=0.0)
    sys = sys.tree_replace({'opt.timestep': 0.001})
    sys = sys.replace(solver_iterations=500)

    state = pipeline.init(sys, init_q, init_qd)
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
    x = state.x

    # compare against generalized step
    state = g_pipeline.init(sys, init_q, init_qd)
    j_g_step = jax.jit(g_pipeline.step)
    j_forward = jax.jit(kinematics.forward)
    for _ in range(1000):
      state = j_g_step(sys, state, jp.zeros(sys.qd_size()))
    x_g, _ = j_forward(sys, state.q, state.qd)

    # trajectories should be close after 1 second of simulation
    np.testing.assert_allclose(x_g.pos, x.pos, rtol=5e-3, atol=5e-3)

  def test_2d_sliding_joint(self):
    # tests launching a capsule at a wall with 2 sliding dofs
    sys = test_utils.load_fixture('double_prismatic.xml')
    sys = sys.replace(
        link=sys.link.replace(
            constraint_vel_damping=150 * jp.ones(sys.num_links())
        )
    )
    sys = sys.tree_replace({'opt.timestep': 0.001})

    qd = jp.zeros(sys.qd_size())
    qd = qd.at[0].set(2.5)
    qd = qd.at[1].set(2.5)

    state = pipeline.init(sys, sys.init_q, qd)
    j_spring_step = jax.jit(pipeline.step)
    states = []
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
      states.append(state)
    x, xd = state.x, state.xd

    # capsule still constrained to the plane and not rotating
    self.assertAlmostEqual(x.pos[0, 2], 0.0, delta=1e-2)
    np.testing.assert_allclose(
        x.rot, jp.array([[1.0, 0.0, 0.0, 0.0]]), atol=1e-3
    )
    # capsule reflects off boundary and is traveling 2.5 m/s without rotating
    np.testing.assert_allclose(
        xd.vel[0], jp.array([-2.5, -2.5, 0.0]), atol=1e-3
    )
    np.testing.assert_allclose(xd.ang[0], jp.array([0.0, 0.0, 0.0]), atol=1e-7)

  def test_3d_sliding_joint(self):
    # tests launching a capsule at a wall with 3 sliding dofs
    sys = test_utils.load_fixture('triple_prismatic.xml')
    sys = sys.tree_replace({'opt.timestep': 0.001})

    qd = jp.zeros(sys.qd_size())
    qd = qd.at[0].set(2.5)
    qd = qd.at[1].set(2.5)
    qd = qd.at[2].set(2.5)

    state = pipeline.init(sys, sys.init_q, qd)
    j_spring_step = jax.jit(pipeline.step)
    states = []
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
      states.append(state)
    x, xd = state.x, state.xd

    # capsule not rotating
    np.testing.assert_allclose(
        x.rot, jp.array([[1.0, 0.0, 0.0, 0.0]]), atol=1e-3
    )
    # capsule reflected off boundary and is traveling 2.5 m/s without rotating
    np.testing.assert_allclose(
        xd.vel[0], jp.array([-2.5, -2.5, -2.5]), atol=1e-3
    )
    np.testing.assert_allclose(xd.ang[0], jp.array([0.0, 0.0, 0.0]), atol=1e-7)

  def test_2d_prismaversal_joint(self):
    # tests a prismatic+universal 2dof joint sliding/rotating into its limits
    sys = test_utils.load_fixture('prismaversal_2dof_joint.xml')
    sys = sys.tree_replace({'opt.timestep': 0.001})

    qd = jp.zeros(sys.qd_size())
    qd = qd.at[0].set(2.5)
    qd = qd.at[1].set(2.5)

    state = pipeline.init(sys, sys.init_q, qd)
    j_spring_step = jax.jit(pipeline.step)
    states = []
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
      states.append(state)

    # reflects off limits and is still traveling 2.5 m/s
    np.testing.assert_allclose(state.qd, jp.array([-2.5, -2.5]), atol=1e-3)
    np.testing.assert_allclose(
        state.xd.ang[0], jp.array([0.0, -2.5, 0.0]), atol=1e-4
    )

  def test_3d_prismaversal_joint(self):
    # tests a prismatic+spherical 3dof joint sliding/rotating into its limits
    sys = test_utils.load_fixture('prismaversal_3dof_joint.xml')
    sys = sys.tree_replace({'opt.timestep': 0.001})

    qd = jp.zeros(sys.qd_size())
    qd = qd.at[0].set(2.5)
    qd = qd.at[1].set(2.5)
    qd = qd.at[2].set(2.5)

    state = pipeline.init(sys, sys.init_q, qd)
    j_spring_step = jax.jit(pipeline.step)
    states = []
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
      states.append(state)

    # reflects off limits and is still traveling 2.5 m/s
    np.testing.assert_allclose(
        state.qd, jp.array([-2.5, -2.5, -2.5]), atol=1e-3
    )
    np.testing.assert_allclose(
        state.xd.ang[0], jp.array([0.0, -2.5, 0.0]), atol=1e-4
    )

  def test_sliding_capsule(self):
    sys = test_utils.load_fixture('capsule.xml')
    sys = sys.tree_replace({'opt.timestep': 0.001})

    qd = jp.zeros(sys.qd_size())
    qd = qd.at[0].set(5.0)

    state = pipeline.init(sys, sys.init_q, qd)
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))
    x, xd = state.x, state.xd

    # capsule slides to a stop
    self.assertAlmostEqual(x.pos[0, 2], 0.25, delta=1e-3)
    np.testing.assert_allclose(
        x.rot, jp.array([[1.0, 0.0, 0.0, 0.0]]), atol=1e-3
    )
    np.testing.assert_allclose(xd.vel, jp.zeros_like(xd.vel), atol=1e-3)
    np.testing.assert_allclose(xd.ang, jp.zeros_like(xd.ang), atol=1e-3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core brax structs and some conversion and slicing functions."""

import copy
from typing import Optional, Sequence, Tuple, Union
import warnings

from brax.v1 import jumpy as jp
from brax.v1 import math
from brax.v1.io import mesh
from brax.v1.physics import config_pb2
from flax import struct


@struct.dataclass
class Q(object):
  """Coordinates: position and rotation.

  Attributes:
    pos: Location of center of mass.
    rot: Rotation about center of mass, represented as a quaternion.
  """
  pos: jp.ndarray
  rot: jp.ndarray

  def __add__(self, o):
    if isinstance(o, P):
      return QP(self.pos, self.rot, o.vel, o.ang)
    elif isinstance(o, Q):
      return Q(self.pos + o.pos, self.rot + o.rot)
    elif isinstance(o, QP):
      return QP(self.pos + o.pos, self.rot + o.rot, o.vel, o.ang)
    else:
      raise ValueError('add only supported for P, Q, QP')


@struct.dataclass
class P(object):
  """Time derivatives: velocity and angular velocity.

  Attributes:
    vel: Velocity.
    ang: Angular velocity about center of mass.
  """
  vel: jp.ndarray
  ang: jp.ndarray

  def __add__(self, o):
    if isinstance(o, P):
      return P(self.vel + o.vel, self.ang + o.ang)
    elif isinstance(o, Q):
      return QP(o.pos, o.rot, self.vel, self.ang)
    elif isinstance(o, QP):
      return QP(o.pos, o.rot, self.vel + o.vel, self.ang + o.ang)
    else:
      raise ValueError('add only supported for P, Q, QP')

  def __mul__(self, o):
    return P(self.vel * o, self.ang * o)


@struct.dataclass
class QP(object):
  """A coordinate and time derivative frame for a brax body.

  Attributes:
    pos: Location of center of mass.
    rot: Rotation about center of mass, represented as a quaternion.
    vel: Velocity.
    ang: Angular velocity about center of mass.
  """
  pos: jp.ndarray
  rot: jp.ndarray
  vel: jp.ndarray
  ang: jp.ndarray

  def __add__(self, o):
    if isinstance(o, P):
      return QP(self.pos, self.rot, self.vel + o.vel, self.ang + o.ang)
    elif isinstance(o, Q):
      return QP(self.pos + o.pos, self.rot + o.rot, self.vel, self.ang)
    elif isinstance(o, QP):
      return QP(self.pos + o.pos, self.rot + o.rot, self.vel + o.vel,
                self.ang + o.ang)
    else:
      raise ValueError('add only supported for P, Q, QP')

  def __mul__(self, o):
    return QP(self.pos * o, self.rot * o, self.vel * o, self.ang * o)

  @classmethod
  def zero(cls, shape=()):
    return cls(
        pos=jp.zeros(shape + (3,)),
        rot=jp.tile(jp.array([1., 0., 0., 0]), reps=shape + (1,)),
        vel=jp.zeros(shape + (3,)),
        ang=jp.zeros(shape + (3,)))

  def to_world(self, rpos: jp.ndarray) -> Tuple[jp.ndarray, jp.ndarray]:
    """Returns world information about a point relative to a part.

    Args:
      rpos: Point relative to center of mass of part.

    Returns:
      A 2-tuple containing:
        * World-space coordinates of rpos
        * World-space velocity of rpos
    """
    rpos_off = math.rotate(rpos, self.rot)
    rvel = jp.cross(self.ang, rpos_off)
    return (self.pos + rpos_off, self.vel + rvel)

  def world_velocity(self, pos: jp.ndarray) -> jp.ndarray:
    """Returns the velocity of the point on a rigidbody in world space.

    Args:
      pos: World space position which to use for velocity calculation.
    """
    return self.vel + jp.cross(self.ang, pos - self.pos)


@struct.dataclass
class Info(object):
  """Auxilliary data calculated during the dynamics of each physics step.

  Attributes:
    contact: External contact forces applied at a step
    joint: Joint constraint forces applied at a step
    actuator: Actuator forces applied at a step
    contact_pos: Contact point position in world coordinates.
    contact_normal: Contact normal.
    contact_penetration: Contact penetration.
  """
  contact: P
  joint: Union[P, Q]
  actuator: P
  contact_pos: jp.ndarray
  contact_normal: jp.ndarray
  contact_penetration: jp.ndarray


def validate_config(
    config: config_pb2.Config,
    resource_paths: Optional[Sequence[str]] = None) -> config_pb2.Config:
  """Validate and normalize config settings for use in systems."""
  config = copy.deepcopy(config)

  if config.dt <= 0:
    raise ValueError('config.dt must be positive')

  if config.substeps == 0:
    config.substeps = 1

  config.solver_scale_collide = config.solver_scale_collide or 1.0

  def find_dupes(objs):
    names = set()
    for obj in objs:
      if obj.name in names:
        raise RuntimeError(f'duplicate name in config: {obj.name}')
      names.add(obj.name)

  find_dupes(config.bodies)
  find_dupes(config.joints)
  find_dupes(config.actuators)
  find_dupes(config.mesh_geometries)

  if config.dynamics_mode == 'legacy_spring':
    if any(j.stiffness == 0 for j in config.joints):
      raise ValueError(
          'joint.stiffness must be >0 when dynamics_mode == legacy_spring')
  elif config.dynamics_mode == 'pbd':
    if any(j.stiffness != 0 for j in config.joints):
      raise ValueError('joint.stiffness is invalid when dynamics_mode == pbd')
    if config.baumgarte_erp:
      raise ValueError('baumgarte_erp is invalid when dynamics_mode == pbd')
  elif any(j.stiffness != 0 for j in config.joints):
    config.dynamics_mode = 'legacy_spring'
    warnings.warn('dynamics_mode not specified, but joint.stiffness >0. '
                  'Setting dynamics_mode="legacy_spring".')
  else:
    config.dynamics_mode = 'pbd'
    warnings.warn(
        'dynamics_mode either not specified or not recognized, defaulting to '
        '"pbd".  If you wish to preserve legacy behavior used in previous '
        'versions of Brax, set dynamics_mode="legacy_spring".'
    )

  # Load the meshes.
  if resource_paths is None:
    resource_paths = ['']
  for i in range(len(config.mesh_geometries)):
    path = config.mesh_geometries[i].path
    if not path:
      continue
    mesh_geom = mesh.load(config.mesh_geometries[i].name, path, resource_paths)
    config.mesh_geometries[i].CopyFrom(mesh_geom)

  # TODO: more config validation

  # reify all frozen dimensions in the system
  allvec = config_pb2.Vector3(x=1.0, y=1.0, z=1.0)
  frozen = config.frozen
  if frozen.all:
    frozen.position.CopyFrom(allvec)
    frozen.rotation.CopyFrom(allvec)
  if all([
      frozen.position.x, frozen.position.y, frozen.position.z,
      frozen.rotation.x, frozen.rotation.y, frozen.rotation.z
  ]):
    config.frozen.all = True
  for b in config.bodies:
    inertia = b.inertia
    if inertia.x == 0 and inertia.y == 0 and inertia.z == 0:
      b.inertia.x, b.inertia.y, b.inertia.z = 1, 1, 1

    b.frozen.position.x = b.frozen.position.x or frozen.position.x
    b.frozen.position.y = b.frozen.position.y or frozen.position.y
    b.frozen.position.z = b.frozen.position.z or frozen.position.z
    b.frozen.rotation.x = b.frozen.rotation.x or frozen.rotation.x
    b.frozen.rotation.y = b.frozen.rotation.y or frozen.rotation.y
    b.frozen.rotation.z = b.frozen.rotation.z or frozen.rotation.z
    if b.frozen.all:
      b.frozen.position.CopyFrom(allvec)
      b.frozen.rotation.CopyFrom(allvec)
    if all([
        b.frozen.position.x, b.frozen.position.y, b.frozen.position.z,
        b.frozen.rotation.x, b.frozen.rotation.y, b.frozen.rotation.z
    ]):
      b.frozen.all = True

    # insert material properties to colliders
    for c in b.colliders:
      if not c.HasField('material'):
        c.material.friction = config.friction
        c.material.elasticity = config.elasticity

  frozen.all = all(b.frozen.all for b in config.bodies)

  return config


def vec_to_arr(vec: config_pb2.Vector3) -> jp.ndarray:
  return jp.array([vec.x, vec.y, vec.z])



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=redefined-builtin
"""Numpy backend for JAX that is called for non-jit/non-jax arrays."""

import builtins
from typing import Any, Callable, List, Optional, Sequence, Tuple, TypeVar, Union
import warnings

import jax
from jax import custom_jvp
from jax import numpy as jnp
import numpy as onp

ndarray = Union[onp.ndarray, jnp.ndarray]  # pylint:disable=invalid-name
tree_map = jax.tree_util.tree_map  # works great with jax or numpy as-is
pi = onp.pi
inf = onp.inf
float32 = onp.float32
int32 = onp.int32

warnings.warn(
    'brax.v1 is deprecated and will be removed in a future release.',
    DeprecationWarning)


def _in_jit() -> bool:
  """Returns true if currently inside a jax.jit call or jit is disabled."""
  if jax.config.jax_disable_jit:
    return True
  return 'DynamicJaxprTrace' in str(
      jax.core.thread_local_state.trace_state.trace_stack
  )


def _which_np(*args):
  checker = lambda a: (
      isinstance(a, (jnp.ndarray, jax.interpreters.batching.BatchTracer)) and
      not isinstance(a, onp.ndarray))
  if builtins.any(jax.tree_util.tree_leaves(tree_map(checker, args))):
    return jnp
  return onp


F = TypeVar('F', bound=Callable)


def vmap(fun: F, include: Optional[Sequence[bool]] = None) -> F:
  """Creates a function which maps ``fun`` over argument axes."""
  if _in_jit():
    in_axes = 0
    if include:
      in_axes = [0 if inc else None for inc in include]
    return jax.vmap(fun, in_axes=in_axes)

  def _batched(*args, include=include):
    if include is not None and len(include) != len(args):
      raise RuntimeError('Len of `args` list must match length of `include`.')

    # by default, vectorize over every arg
    if include is None:
      include = [True for _ in args]

    # determine number of parallel evaluations to unroll into serial evals
    batch_size = None
    for a, inc in zip(args, include):
      if inc:
        flat_args, _ = jax.tree_util.tree_flatten(a)
        batch_size = flat_args[0].shape[0]
        break

    # rebuild b_args for each serial evaluation
    rets = []
    for b_idx in range(batch_size):
      b_args = []
      for a, inc in zip(args, include):
        if inc:
          b_args.append(take(a, b_idx))  # pytype: disable=wrong-arg-types  # jax-ndarray
        else:
          b_args.append(a)
      rets.append(fun(*b_args))

    np = _which_np(*rets)
    return jax.tree_util.tree_map(lambda *x: np.stack(x), *rets)

  return _batched


Carry = TypeVar('Carry')
X = TypeVar('X')
Y = TypeVar('Y')


def scan(f: Callable[[Carry, X], Tuple[Carry, Y]],
         init: Carry,
         xs: X,
         length: Optional[int] = None,
         reverse: bool = False,
         unroll: int = 1) -> Tuple[Carry, Y]:
  """Scan a function over leading array axes while carrying along state."""
  if _in_jit():
    return jax.lax.scan(f, init, xs, length, reverse, unroll)
  else:
    xs_flat, xs_tree = jax.tree_util.tree_flatten(xs)
    carry = init
    ys = []
    maybe_reversed = reversed if reverse else lambda x: x
    for i in maybe_reversed(range(length)):
      xs_slice = [x[i] for x in xs_flat]
      carry, y = f(carry, jax.tree_util.tree_unflatten(xs_tree, xs_slice))
      ys.append(y)
    mry = maybe_reversed(ys)
    np = _which_np(*mry)
    stacked_y = jax.tree_util.tree_map(lambda *y: np.stack(y), *mry)
    return carry, stacked_y


def while_loop(cond_fun: Callable[[X], Any], body_fun: Callable[[X], X],
               init_val: X) -> X:
  """Call body_fun while cond_fun is true, starting with init_val."""
  if _in_jit():
    return jax.lax.while_loop(cond_fun, body_fun, init_val)
  else:
    val = init_val
    while cond_fun(val):
      val = body_fun(val)
    return val


def fori_loop(lower: int, upper: int, body_fun: Callable[[int, X], X],
              init_val: X) -> X:
  """Call body_fun over range from lower to upper, starting with init_val."""
  if _in_jit():
    return jax.lax.fori_loop(lower, upper, body_fun, init_val)
  else:
    val = init_val
    for i in range(lower, upper):
      val = body_fun(i, val)
    return val


def take(
    tree: Any, i: Union[int, ndarray, Sequence[int]], axis: int = 0
) -> Any:
  """Returns tree sliced by i."""
  np = _which_np(i)
  if isinstance(i, list) or isinstance(i, tuple):
    i = np.array(i, dtype=int)
  return jax.tree_util.tree_map(lambda x: np.take(x, i, axis=axis, mode='clip'),
                                tree)


def norm(x: ndarray,
         axis: Optional[Union[Tuple[int, ...], int]] = None) -> ndarray:
  """Returns the array norm."""
  return _which_np(x, axis).linalg.norm(x, axis=axis)


def index_update(x: ndarray, idx: ndarray, y: ndarray) -> ndarray:
  """Pure equivalent of x[idx] = y."""
  if _which_np(x, idx, y) is jnp:
    return jnp.array(x).at[idx].set(jnp.array(y))
  x = onp.copy(x)
  x[idx] = y
  return x


def safe_norm(x: ndarray,
              axis: Optional[Union[Tuple[int, ...], int]] = None) -> ndarray:
  """Calculates a linalg.norm(x) that's safe for gradients at x=0.

  Avoids a poorly defined gradient for jnp.linal.norm(0) see
  https://github.com/google/jax/issues/3058 for details
  Args:
    x: A jnp.array
    axis: The axis along which to compute the norm

  Returns:
    Norm of the array x.
  """
  np = _which_np(x)
  if np is jnp:
    is_zero = jnp.allclose(x, 0.)
    # temporarily swap x with ones if is_zero, then swap back
    x = jnp.where(is_zero, jnp.ones_like(x), x)
    n = jnp.linalg.norm(x, axis=axis)
    n = jnp.where(is_zero, 0., n)
  else:
    n = onp.linalg.norm(x, axis=axis)
  return n


def expand_dims(x: ndarray,
                axis: Union[Tuple[int, ...], int] = 0) -> ndarray:
  """Increases batch dimensionality along axis."""

  return _which_np(x).expand_dims(x, axis=axis)


def any(a: ndarray, axis: Optional[int] = None) -> ndarray:
  """Test whether any array element along a given axis evaluates to True."""
  return _which_np(a).any(a, axis=axis)


def all(a: ndarray, axis: Optional[int] = None) -> ndarray:
  """Test whether all array elements along a given axis evaluate to True."""
  return _which_np(a).all(a, axis=axis)


def mean(a: ndarray, axis: Optional[int] = None) -> ndarray:
  """Compute the arithmetic mean along the specified axis."""
  return _which_np(a).mean(a, axis=axis)


def var(a: ndarray, axis: Optional[int] = None) -> ndarray:
  """Compute the variance along the specified axis."""
  return _which_np(a).var(a, axis=axis)


def arange(start: int, stop: int) -> ndarray:
  """Return evenly spaced values within a given interval."""
  return _which_np().arange(start, stop)


def dot(x: ndarray, y: ndarray) -> ndarray:
  """Returns dot product of two arrays."""
  return _which_np(x, y).dot(x, y)


def outer(a: ndarray, b: ndarray) -> ndarray:
  """Compute the outer product of two vectors."""
  return _which_np(a, b).outer(a, b)


def matmul(x1: ndarray, x2: ndarray) -> ndarray:
  """Matrix product of two arrays."""
  return _which_np(x1, x2).matmul(x1, x2)


def inv(a: ndarray) -> ndarray:
  """Compute the (multiplicative) inverse of a matrix."""
  return _which_np(a).linalg.inv(a)


def roll(x: ndarray, shift, axis=None) -> ndarray:
  """Rolls array elements along a given axis."""
  return _which_np(x).roll(x, shift, axis=axis)


def square(x: ndarray) -> ndarray:
  """Return the element-wise square of the input."""
  return _which_np(x).square(x)


def tile(x: ndarray, reps: Union[Tuple[int, ...], int]) -> ndarray:
  """Construct an array by repeating A the number of times given by reps."""
  return _which_np(x).tile(x, reps)


def repeat(a: ndarray, repeats: Union[int, ndarray], *args,
           **kwargs) -> ndarray:
  """Repeat elements of an array."""
  return _which_np(a, repeats).repeat(a, repeats=repeats, *args, **kwargs)


def floor(x: ndarray) -> ndarray:
  """Returns the floor of the input, element-wise.."""
  return _which_np(x).floor(x)


def cross(x: ndarray, y: ndarray) -> ndarray:
  """Returns cross product of two arrays."""
  return _which_np(x, y).cross(x, y)


def sin(angle: ndarray) -> ndarray:
  """Returns trigonometric sine, element-wise."""
  return _which_np(angle).sin(angle)


def cos(angle: ndarray) -> ndarray:
  """Returns trigonometric cosine, element-wise."""
  return _which_np(angle).cos(angle)


def arctan2(x1: ndarray, x2: ndarray) -> ndarray:
  """Returns element-wise arc tangent of x1/x2 choosing the quadrant correctly."""
  return _which_np(x1, x2).arctan2(x1, x2)


def arctanh(x: ndarray) -> ndarray:
  """Returns element-wise arctanh of x."""
  return _which_np(x).arctanh(x)


def tanh(x: ndarray) -> ndarray:
  """Returns element-wise tanh of x."""
  return _which_np(x).tanh(x)


def arccos(x: ndarray) -> ndarray:
  """Trigonometric inverse cosine, element-wise."""
  return _which_np(x).arccos(x)


@custom_jvp
def safe_arccos(x: ndarray) -> ndarray:
  """Trigonometric inverse cosine, element-wise with safety clipping in grad."""
  return _which_np(x).arccos(x)


@safe_arccos.defjvp
def _safe_arccos_jvp(primal, tangent):
  x, = primal
  x_dot, = tangent
  primal_out = safe_arccos(x)
  tangent_out = -x_dot / sqrt(1. - clip(x, -1 + 1e-7, 1 - 1e-7)**2.)  # pytype: disable=wrong-arg-types  # jax-ndarray
  return primal_out, tangent_out


def arcsin(x: ndarray) -> ndarray:
  """Trigonometric inverse sine, element-wise."""
  return _which_np(x).arcsin(x)


@custom_jvp
def safe_arcsin(x: ndarray) -> ndarray:
  """Trigonometric inverse sine, element-wise with safety clipping in grad."""
  return _which_np(x).arcsin(x)


@safe_arcsin.defjvp
def _safe_arcsin_jvp(primal, tangent):
  x, = primal
  x_dot, = tangent
  primal_out = safe_arccos(x)
  tangent_out = x_dot / sqrt(1. - clip(x, -1 + 1e-7, 1 - 1e-7)**2.)  # pytype: disable=wrong-arg-types  # jax-ndarray
  return primal_out, tangent_out


def logical_not(x: ndarray) -> ndarray:
  """Returns the truth value of NOT x element-wise."""
  return _which_np(x).logical_not(x)


def logical_and(x1: ndarray, x2: ndarray) -> ndarray:
  """Returns the truth value of x1 AND x2 element-wise."""
  return _which_np(x1, x2).logical_and(x1, x2)


def logical_or(x1: ndarray, x2: ndarray) -> ndarray:
  """Returns the truth value of x1 OR x2 element-wise."""
  return _which_np(x1, x2).logical_or(x1, x2)


def multiply(x1: ndarray, x2: ndarray) -> ndarray:
  """Multiply arguments element-wise."""
  return _which_np(x1, x2).multiply(x1, x2)


def minimum(x1: ndarray, x2: ndarray) -> ndarray:
  """Element-wise minimum of array elements."""
  return _which_np(x1, x2).minimum(x1, x2)


def maximum(x1: ndarray, x2: ndarray) -> ndarray:
  """Element-wise maximum of array elements."""
  return _which_np(x1, x2).maximum(x1, x2)


def amin(x: ndarray, *args, **kwargs) -> ndarray:
  """Returns the minimum along a given axis."""
  return _which_np(x).amin(x, *args, **kwargs)


def amax(x: ndarray, *args, **kwargs) -> ndarray:
  """Returns the maximum along a given axis."""
  return _which_np(x).amax(x, *args, **kwargs)


def argmin(x: ndarray, *args, **kwargs) -> ndarray:
  """Returns the argmin along a given axis."""
  return _which_np(x).argmin(x, *args, **kwargs)


def argmax(x: ndarray, *args, **kwargs) -> ndarray:
  """Returns the argmax along a given axis."""
  return _which_np(x).argmax(x, *args, **kwargs)


def exp(x: ndarray) -> ndarray:
  """Returns the exponential of all elements in the input array."""
  return _which_np(x).exp(x)


def sign(x: ndarray) -> ndarray:
  """Returns an element-wise indication of the sign of a number."""
  return _which_np(x).sign(x)


def sum(a: ndarray, axis: Optional[int] = None):
  """Returns sum of array elements over a given axis."""
  return _which_np(a).sum(a, axis=axis)


def random_prngkey(seed: int) -> ndarray:
  """Returns a PRNG key given a seed."""
  if _which_np() is jnp:
    return jax.random.PRNGKey(seed)
  else:
    rng = onp.random.default_rng(seed)
    return rng.integers(low=0, high=2**32, dtype='uint32', size=2)


def random_uniform(rng: ndarray,
                   shape: Tuple[int, ...] = (),
                   low: Optional[float] = 0.0,
                   high: Optional[float] = 1.0) -> ndarray:
  """Sample uniform random values in [low, high) with given shape/dtype."""
  if _which_np(rng) is jnp:
    return jax.random.uniform(rng, shape=shape, minval=low, maxval=high)
  else:
    return onp.random.default_rng(rng).uniform(size=shape, low=low, high=high)


def random_split(rng: ndarray, num: int = 2) -> ndarray:
  """Splits a PRNG key into num new keys by adding a leading axis."""
  if _which_np(rng) is jnp:
    return jax.random.split(rng, num=num)
  else:
    rng = onp.random.default_rng(rng)
    return rng.integers(low=0, high=2**32, dtype='uint32', size=(num, 2))


def randint(rng: ndarray,
            shape: Tuple[int, ...] = (),
            low: Optional[int] = 0,
            high: Optional[int] = 1) -> ndarray:
  """Sample integers in [low, high) with given shape."""
  if _which_np(rng) is jnp:
    return jax.random.randint(rng, shape=shape, minval=low, maxval=high)
  else:
    return onp.random.default_rng(rng).integers(low=low, high=high, size=shape)


def choice(rng: ndarray,
           a: Union[int, Any],
           shape: Tuple[int, ...] = (),
           replace: bool = True,
           p: Optional[Any] = None,
           axis: int = 0) -> ndarray:
  """Generate sample(s) from given array."""
  if _which_np(rng) is jnp:
    return jax.random.choice(
        rng, a, shape=shape, replace=replace, p=p, axis=axis)
  else:
    return onp.random.default_rng(rng).choice(
        a, size=shape, replace=replace, p=p, axis=axis)


def segment_sum(data: ndarray,
                segment_ids: ndarray,
                num_segments: Optional[int] = None) -> ndarray:
  """Computes the sum within segments of an array."""
  if _which_np(data, segment_ids) is jnp:
    s = jax.ops.segment_sum(data, segment_ids, num_segments)
  else:
    if num_segments is None:
      num_segments = onp.amax(segment_ids) + 1
    s = onp.zeros((num_segments,) + data.shape[1:])
    onp.add.at(s, segment_ids, data)
  return s


def top_k(operand: ndarray, k: int) -> ndarray:
  """Returns the ordered top k values and their indices along the last axis of operand."""
  if _which_np(operand) is jnp:
    return jax.lax.top_k(operand, k)  # pytype: disable=bad-return-type  # jax-ndarray
  else:
    top_ind = onp.argpartition(operand, -k)[-k:]
    sorted_ind = top_ind[onp.argsort(-operand[top_ind])]
    return operand[sorted_ind], sorted_ind  # pytype: disable=bad-return-type  # jax-ndarray


def stack(x: List[ndarray], axis=0) -> ndarray:
  """Join a sequence of arrays along a new axis."""
  return _which_np(*x).stack(x, axis=axis)


def concatenate(x: Sequence[ndarray], axis=0) -> ndarray:
  """Join a sequence of arrays along an existing axis."""
  return _which_np(*x).concatenate(x, axis=axis)


def sqrt(x: ndarray) -> ndarray:
  """Returns the non-negative square-root of an array, element-wise."""
  return _which_np(x).sqrt(x)


def where(condition: jax.typing.ArrayLike, x: jax.typing.ArrayLike,
          y: jax.typing.ArrayLike) -> ndarray:
  """Return elements chosen from `x` or `y` depending on `condition`."""
  return _which_np(condition, x, y).where(condition, x, y)


def cond(pred, true_fun: Callable[..., bool], false_fun: Callable[..., bool],
         *operands: Any):
  """Conditionally apply true_fun or false_fun to operands."""
  if _in_jit():
    return jax.lax.cond(pred, true_fun, false_fun, *operands)
  else:
    if pred:
      return true_fun(operands)
    else:
      return false_fun(operands)


def diag(v: ndarray, k: int = 0) -> ndarray:
  """Extract a diagonal or construct a diagonal array."""
  return _which_np(v).diag(v, k)


def clip(a: ndarray, a_min: ndarray, a_max: ndarray) -> ndarray:
  """Clip (limit) the values in an array."""
  return _which_np(a, a_min, a_max).clip(a, a_min, a_max)


def eye(n: int) -> ndarray:
  """Return a 2-D array with ones on the diagonal and zeros elsewhere."""
  return _which_np().eye(n)


def zeros(shape, dtype=float) -> ndarray:
  """Return a new array of given shape and type, filled with zeros."""
  return _which_np().zeros(shape, dtype=dtype)


def zeros_like(a: ndarray) -> ndarray:
  """Return an array of zeros with the same shape and type as a given array."""
  return _which_np(a).zeros_like(a)


def ones(shape, dtype=float) -> ndarray:
  """Return a new array of given shape and type, filled with ones."""
  return _which_np().ones(shape, dtype=dtype)


def ones_like(a: ndarray) -> ndarray:
  """Return an array of ones with the same shape and type as a given array."""
  return _which_np(a).ones_like(a)


def reshape(a: ndarray, newshape: Union[Tuple[int, ...], int]) -> ndarray:
  """Gives a new shape to an array without changing its data."""
  return _which_np(a).reshape(a, newshape)


def atleast_1d(*arys) -> ndarray:
  """Ensure arrays are all at least 1d (dimensions added to beginning)."""
  return _which_np(*arys).atleast_1d(*arys)


def atleast_2d(*arys) -> ndarray:
  """Ensure arrays are all at least 2d (dimensions added to beginning)."""
  return _which_np(*arys).atleast_2d(*arys)


def atleast_3d(*arys) -> ndarray:
  """Ensure arrays are all at least 3d (dimensions added to beginning)."""
  return _which_np(*arys).atleast_3d(*arys)


def array(object: Any, dtype=None) -> ndarray:
  """Creates an array given a list."""
  try:
    np = _which_np(*object)
  except TypeError:
    np = _which_np(object)  # object is not iterable (e.g. primitive type)
  return np.array(object, dtype)


def abs(a: ndarray) -> ndarray:
  """Calculate the absolute value element-wise."""
  return _which_np(a).abs(a)


def meshgrid(*xi,
             copy: bool = True,
             sparse: bool = False,
             indexing: str = 'xy') -> ndarray:
  """Create N-D coordinate matrices from 1D coordinate vectors."""
  if _which_np(xi[0]) is jnp:
    return jnp.meshgrid(*xi, copy=copy, sparse=sparse, indexing=indexing)  # pytype: disable=bad-return-type  # jnp-type
  return onp.meshgrid(*xi, copy=copy, sparse=sparse, indexing=indexing)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Saves a system config and trajectory as json."""

import json
from typing import List, Text, Tuple

from brax.base import State, System
from etils import epath
import jax
import jax.numpy as jp
import mujoco
import numpy as np


# State attributes needed for the visualizer.
_STATE_ATTR = ['x', 'contact']

# Fields that get json encoded.
_ENCODE_FIELDS = [
    'contact',
    'opt',
    'timestep',
    'face',
    'size',
    'link_idx',
    'link_names',
    'name',
    'dist',
    'pos',
    'rgba',
    'rot',
    'states',
    'transform',
    'vert',
    'x',
    'xd',
]

_GEOM_TYPE_NAMES = {
    0: 'Plane',
    1: 'HeightMap',
    2: 'Sphere',
    3: 'Capsule',
    5: 'Cylinder',
    6: 'Box',
    7: 'Mesh',
}


def _to_dict(obj):
  """Converts python object to a json encodeable object."""
  if isinstance(obj, list) or isinstance(obj, tuple):
    return [_to_dict(s) for s in obj]
  if isinstance(obj, dict):
    return {k: _to_dict(v) for k, v in obj.items() if k in _ENCODE_FIELDS}
  if isinstance(obj, jax.Array):
    return _to_dict(obj.tolist())
  if hasattr(obj, '__dict__'):
    d = dict(obj.__dict__)
    d['name'] = obj.__class__.__name__
    return _to_dict(d)
  if isinstance(obj, np.ndarray):
    return _to_dict(obj.tolist())
  if isinstance(obj, np.floating):
    return float(obj)
  if isinstance(obj, np.integer):
    return int(obj)
  if isinstance(obj, float) and (np.isnan(obj) or np.isinf(obj)):
    return str(obj)
  return obj


def _compress_contact(states: State) -> State:
  """Reduces the number of contacts based on penetration > 0."""
  if states.contact is None or states.contact.pos.shape[0] == 0:
    return states

  contact_mask = states.contact.dist < 0
  n_contact = contact_mask.sum(axis=1).max()

  def pad(arr, n):
    r = jp.zeros(n)
    if len(arr.shape) > 1:
      r = jp.zeros((n, *arr.shape[1:]))
    r = r.at[: arr.shape[0], ...].set(arr)
    return r

  def compress(contact, i):
    return jax.tree.map(
        lambda x: pad(x[contact_mask[i]], n_contact), contact.take(i)
    )

  c = [compress(states.contact, i) for i in range(states.x.pos.shape[0])]
  return states.replace(contact=jax.tree.map(lambda *x: jp.stack(x), *c))


def _get_mesh(mj: mujoco.MjModel, i: int) -> Tuple[np.ndarray, np.ndarray]:
  """Gets mesh from mj at index i."""
  last = (i + 1) >= mj.nmesh
  face_start = mj.mesh_faceadr[i]
  face_end = mj.mesh_faceadr[i + 1] if not last else mj.mesh_face.shape[0]
  face = mj.mesh_face[face_start:face_end]

  vert_start = mj.mesh_vertadr[i]
  vert_end = mj.mesh_vertadr[i + 1] if not last else mj.mesh_vert.shape[0]
  vert = mj.mesh_vert[vert_start:vert_end]

  return vert, face


def dumps(sys: System, states: List[State]) -> Text:
  """Creates a json string of the system config.

  Args:
    sys: brax System object
    states: list of brax system states

  Returns:
    string containing json dump of system and states

  Raises:
    RuntimeError: if states have invalid shape
  """
  if any((len(s.x.pos.shape), len(s.x.rot.shape)) != (2, 2) for s in states):
    pos_shape = max(len(s.x.pos.shape) for s in states)
    rot_shape = max(len(s.x.rot.shape) for s in states)
    raise RuntimeError(
        'Expected state.x position and rotation to have 2 shape dimensions but '
        f'received len(pos.shape)={pos_shape} and len(rot.shape)={rot_shape}'
    )

  d = _to_dict(sys)

  # TODO: move the manipulations below to javascript

  # fill in empty link names
  link_names = [n or f'link {i}' for i, n in enumerate(sys.link_names)]
  link_names += ['world']

  # unpack geoms into a dict for the visualizer
  link_geoms = {}
  for id_ in range(sys.ngeom):
    link_idx = sys.geom_bodyid[id_] - 1

    rgba = sys.geom_rgba[id_]
    if (rgba == [0.5, 0.5, 0.5, 1.0]).all():
      # convert the default mjcf color to brax default color
      rgba = np.array([0.4, 0.33, 0.26, 1.0])

    geom = {
        'name': _GEOM_TYPE_NAMES[sys.geom_type[id_]],
        'link_idx': link_idx,
        'pos': sys.geom_pos[id_],
        'rot': sys.geom_quat[id_],
        'rgba': rgba,
        'size': sys.geom_size[id_],
    }

    if geom['name'] == 'Mesh':
      vert, face = _get_mesh(sys.mj_model, sys.geom_dataid[id_])
      geom['vert'] = vert
      geom['face'] = face

    link_geoms.setdefault(link_names[link_idx], []).append(_to_dict(geom))

  d['geoms'] = link_geoms

  # stack states for the viewer
  states = jax.tree.map(lambda *x: jp.stack(x), *states)
  states = _compress_contact(states)

  states = _to_dict(states)
  d['states'] = {k: states[k] for k in _STATE_ATTR}

  return json.dumps(d)


def save(path: str, sys: System, states: List[State]):
  """Saves a system config and trajectory as json."""
  with epath.Path(path).open('w') as fout:
    fout.write(dumps(sys, states))



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for joints."""

from absl.testing import absltest
from brax import test_utils
from brax.positional import joints
from brax.positional import pipeline
from jax import numpy as jp
import numpy as np


class JointsTest(absltest.TestCase):

  def test_sphericalize_humanoid(self):
    sys = test_utils.load_fixture('humanoid.xml')
    qd = jp.zeros(sys.qd_size())
    state = pipeline.init(sys, sys.init_q, qd)

    sys_data = joints._sphericalize(sys, state.x)

    # check lower limits
    np.testing.assert_array_almost_equal(
        sys_data[0][0],
        jp.array([
            [-jp.inf, -jp.inf, -jp.inf],
            [-0.7853982, -1.3089969, 0.0],
            [-0.61086524, 0.0, 0.0],
            [-0.43633232, -1.0471976, -1.9198622],
            [-2.7925267, 0.0, 0.0],
            [-0.43633232, -1.0471976, -1.9198622],
            [-2.7925267, 0.0, 0.0],
            [-1.4835298, -1.4835298, 0.0],
            [-1.5707964, 0.0, 0.0],
            [-1.0471976, -1.0471976, 0.0],
            [-1.5707964, 0.0, 0.0],
        ]),
    )

    # check upper limits
    np.testing.assert_array_almost_equal(
        sys_data[0][1],
        jp.array([
            [jp.inf, jp.inf, jp.inf],
            [0.7853982, 0.5235988, 0.0],
            [0.61086524, 0.0, 0.0],
            [0.08726646, 0.61086524, 0.34906584],
            [-0.03490658, 0.0, 0.0],
            [0.08726646, 0.61086524, 0.34906584],
            [-0.03490658, 0.0, 0.0],
            [1.0471976, 1.0471976, 0.0],
            [0.87266463, 0.0, 0.0],
            [1.4835298, 1.4835298, 0.0],
            [0.87266463, 0.0, 0.0],
        ]),
    )

    # check that a 1-dof link motion is padded correctly
    # check that base dof is the same
    np.testing.assert_array_almost_equal(
        sys_data[1].ang[2][0], sys.dof.motion.ang[8]
    )

    # check padding
    np.testing.assert_array_almost_equal(
        sys_data[1].ang[2],
        jp.array([[1.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]),
    )

    # check that a 2-dof link motion is padded correctly
    # check base dof is the same
    np.testing.assert_array_almost_equal(
        sys_data[1].ang[1][0:2], sys.dof.motion.ang[6:8]
    )

    # check padding
    np.testing.assert_array_almost_equal(
        sys_data[1].ang[1],
        jp.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]]),
    )

    # check link parities
    np.testing.assert_array_almost_equal(
        sys_data[3], [1.0, 1.0, 1.0, -1.0, 1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    )

    # spot check parity for first 3-dof link (corresponds to fourth idx above)
    three_dof_link = sys.dof.motion.ang[sys.qd_idx('3')[0:3]]
    np.testing.assert_equal(
        np.dot(
            np.cross(three_dof_link[0], three_dof_link[1]), three_dof_link[2]
        ),
        -1.0,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functions for scanning pytrees given an order defined by a system."""
from typing import Callable, Sequence, TypeVar

from brax.base import QD_WIDTHS, Q_WIDTHS, System
import jax
from jax import numpy as jp

L = TypeVar('L')
Q = TypeVar('Q')
D = TypeVar('D')
Y = TypeVar('Y')


def _take(obj: Y, idxs: Sequence[int]) -> Y:
  """Takes idxs on any pytree given to it.

  XLA executes x[jp.array([1, 2, 3])] slower than x[1:4], so we detect when
  take indices are contiguous, and convert them to slices.

  Args:
    obj: an input pytree
    idxs: indices to take

  Returns:
    obj pytree with leaves taken by idxs
  """

  def take(x):
    if idxs == list(range(idxs[0], idxs[-1] + 1)):
      x = x[idxs[0] : idxs[-1] + 1]
    else:
      x = x.take(jp.array(idxs), axis=0, mode='wrap')
    return x

  return jax.tree.map(take, obj)


def tree(
    sys: System,
    f: Callable[..., Y],
    in_types: str,
    *args,
    reverse: bool = False,
) -> Y:
  r"""Scan a function over each level of a system while carrying along state.

  f is called once for each level of a system tree.  running `tree` on tall tree
  layouts (like long kinematic chains/ropes) may produce long unrolls that are
  slow to jit.

  Args:
    sys: a brax system defining the scan order
    f: a function to be scanned with the following type signature:\
        def f(y, *args) -> y
       where
         ``y`` is the carry value and return value, in link parent order
         ``*args`` are input arguments with types matching ``in_types``
    in_types: string specifying the type of each input arg:
        'l' is an input to be split according to link ranges
        'q' is an input to be split according to q ranges
        'd' is an input to be split according to qd ranges
    *args: the input arguments corresponding to ``in_types``
    reverse: if True, scans up the system tree from leaves to root, otherwise
      root to leaves

  Returns:
    The stacked outputs of ``f``, matching the system link order.
  """
  if len(args) != len(in_types):
    return ValueError('len(args) must match len(in_types)')

  depth_fn = lambda i, p=sys.link_parents: p[i] + 1 and 1 + depth_fn(p[i])
  q_idx, qd_idx, depth_idxs = 0, 0, []
  for i, t in enumerate(sys.link_types):
    depth = depth_fn(i)
    while depth >= len(depth_idxs):
      depth_idxs.append({'l': [], 'q': [], 'd': []})
    depth_idxs[depth]['l'].append(i)
    depth_idxs[depth]['q'].extend(range(q_idx, q_idx + Q_WIDTHS[t]))
    depth_idxs[depth]['d'].extend(range(qd_idx, qd_idx + QD_WIDTHS[t]))
    q_idx, qd_idx = q_idx + Q_WIDTHS[t], qd_idx + QD_WIDTHS[t]
  y, ys = None, []

  if reverse:
    for depth in range(len(depth_idxs) - 1, -1, -1):
      in_args = [_take(a, depth_idxs[depth][t]) for a, t in zip(args, in_types)]

      if y is not None:
        link_idxs = depth_idxs[depth]['l']
        parent_idxs = [sys.link_parents[i] for i in depth_idxs[depth + 1]['l']]
        parent_map = [link_idxs.index(p) for p in parent_idxs]

        def index_sum(x, b=(len(link_idxs),), p=jp.array(parent_map)):
          return jp.zeros(b + x.shape[1:]).at[p].add(x)

        y = jax.tree.map(index_sum, y)

      y = f(y, *in_args)
      ys.insert(0, y)
  else:
    for depth in range(len(depth_idxs)):
      in_args = [_take(a, depth_idxs[depth][t]) for a, t in zip(args, in_types)]

      if y is not None:
        parent_idxs = [sys.link_parents[i] for i in depth_idxs[depth]['l']]
        parent_map = [depth_idxs[depth - 1]['l'].index(p) for p in parent_idxs]
        y = _take(y, parent_map)

      y = f(y, *in_args)
      ys.append(y)

  y = jax.tree.map(lambda *x: jp.concatenate(x), *ys)

  # we concatenated results out of order, so put back in order if needed
  order = sum([d['l'] for d in depth_idxs], [])
  if order != list(range(len(order))):
    y = _take(y, [order.index(i) for i in range(len(order))])

  return y


def link_types(
    sys: System, f: Callable[..., Y], in_types: str, out_types: str, *args
) -> Y:
  r"""Scan a function over System link type ranges.

  Args:
    sys: system defining the kinematic tree and other properties
    f: a function to be scanned with the following type signature:\
          def f(typ, *args) -> y
        where
          ``typ`` is the actuator, link type string
          ``*args`` are input arguments with types matching ``in_types``
          ``y`` is an output arguments with types matching ``out_type``
    in_types: string specifying the type of each input arg:
        'l' is an input to be split according to link ranges
        'q' is an input to be split according to q ranges
        'd' is an input to be split according to qd ranges
    out_types: string specifying the types of the outputs
    *args: the input arguments corresponding to ``in_types``

  Returns:
    The stacked outputs of ``f`` matching the system link order.
  """
  q_idx, qd_idx, typ_order_idxs = 0, 0, []
  typ_order = sorted(set(sys.link_types), key=sys.link_types.find)
  for i, t in enumerate(sys.link_types):
    order = typ_order.index(t)
    while order >= len(typ_order_idxs):
      typ_order_idxs.append({'l': [], 'q': [], 'd': []})
    typ_order_idxs[order]['l'].append(i)
    typ_order_idxs[order]['q'].extend(range(q_idx, q_idx + Q_WIDTHS[t]))
    typ_order_idxs[order]['d'].extend(range(qd_idx, qd_idx + QD_WIDTHS[t]))
    q_idx, qd_idx = q_idx + Q_WIDTHS[t], qd_idx + QD_WIDTHS[t]

  ys = []

  for typ, typ_idxs in zip(typ_order, typ_order_idxs):
    in_args = [_take(a, typ_idxs[t]) for a, t in zip(args, in_types)]
    ys.append(f(typ, *in_args))

  y = jax.tree.map(lambda *x: jp.concatenate(x), *ys)

  # we concatenated results out of order, so put back in order if needed

  ys = [y] if len(out_types) == 1 else y
  out_ys = []
  for idxs, ot in enumerate(out_types):
    order = sum([t[ot] for t in typ_order_idxs], [])
    if order != list(range(len(order))):
      out_ys.append(
          _take(ys[idxs], [order.index(i) for i in range(len(order))])
      )
    else:
      out_ys.append(ys[idxs])
  y = out_ys[0] if len(out_types) == 1 else out_ys

  return y



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functions for applying actuators to a physics pipeline."""

from brax.base import System
import jax
from jax import numpy as jp


def to_tau(
    sys: System, act: jax.Array, q: jax.Array, qd: jax.Array
) -> jax.Array:
  """Convert actuator to a joint force tau.

  Args:
    sys: system defining the kinematic tree and other properties
    act: (act_size,) actuator force input vector
    q: joint position vector
    qd: joint velocity vector

  Returns:
    tau: (qd_size,) vector of joint forces
  """
  if sys.act_size() == 0:
    return jp.zeros(sys.qd_size())

  ctrl_range = sys.actuator.ctrl_range
  force_range = sys.actuator.force_range

  q, qd = q[sys.actuator.q_id], qd[sys.actuator.qd_id]
  act = jp.clip(act, ctrl_range[:, 0], ctrl_range[:, 1])
  # See https://github.com/deepmind/mujoco/discussions/754 for why gear is
  # used for the bias term.
  bias = sys.actuator.gear * (
      q * sys.actuator.bias_q + qd * sys.actuator.bias_qd
  )

  force = sys.actuator.gain * act + bias
  force = jp.clip(force, force_range[:, 0], force_range[:, 1])

  force *= sys.actuator.gear
  tau = jp.zeros(sys.qd_size()).at[sys.actuator.qd_id].add(force)

  return tau
